# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

from typing import Tuple

from attrs import define, field
import numpy as np

from floris.type_dec import floris_array_converter, NDArrayFloat


def pshape(array: np.ndarray, label: str = ""):
    print(label, np.shape(array))


@define
class Vec3:
    """
    Contains 3-component vector information. All arithmetic operators are
    set so that Vec3 objects can operate on and with each other directly.

    Args:
        components (list(numeric, numeric, numeric), numeric): All three vector
            components.
        string_format (str, optional): Format to use in the
            overloaded __str__ function. Defaults to None.
    """
    components: NDArrayFloat = field(converter=floris_array_converter)
    # NOTE: this does not convert elements to float if they are given as int. Is this ok?

    @components.validator
    def _check_components(self, attribute, value) -> None:        
        if np.ndim(value) > 1:
            raise ValueError(f"Vec3 must contain exactly 1 dimension, {np.ndim(value)} were given.")
        if np.size(value) != 3:
            raise ValueError(f"Vec3 must contain exactly 3 components, {np.size(value)} were given.")

    def __add__(self, arg):
        if type(arg) is Vec3:
            return Vec3(self.components + arg.components)
        elif type(arg) is int or type(arg) is float:
            return Vec3(self.components + arg)
        else:
            raise ValueError

    def __sub__(self, arg):
        if type(arg) is Vec3:
            return Vec3(self.components - arg.components)
        elif type(arg) is int or type(arg) is float:
            return Vec3(self.components - arg)
        else:
            raise ValueError

    def __mul__(self, arg):
        if type(arg) is Vec3:
            return Vec3(self.components * arg.components)
        elif type(arg) is int or type(arg) is float:
            return Vec3(self.components * arg)
        else:
            raise ValueError

    def __truediv__(self, arg):
        if type(arg) is Vec3:
            return Vec3(self.components / arg.components)
        elif type(arg) is int or type(arg) is float:
            return Vec3(self.components / arg)
        else:
            raise ValueError

    def __eq__(self, arg):
        return False not in np.isclose([self.x1, self.x2, self.x3], [arg.x1, arg.x2, arg.x3])

    def __hash__(self):
        return hash((self.x1, self.x2, self.x3))

    @property
    def x1(self):
        return self.components[0]

    @x1.setter
    def x1(self, value):
        self.components[0] = float(value)

    @property
    def x2(self):
        return self.components[1]

    @x2.setter
    def x2(self, value):
        self.components[1] = float(value)

    @property
    def x3(self):
        return self.components[2]

    @x3.setter
    def x3(self, value):
        self.components[2] = float(value)

    @property
    def elements(self) -> Tuple[float, float, float]:
        # TODO: replace references to elements with components
        # and remove this @property
        return self.components


def cosd(angle):
    """
    Cosine of an angle with the angle given in degrees.

    Args:
        angle (float): Angle in degrees.

    Returns:
        float
    """
    return np.cos(np.radians(angle))


def sind(angle):
    """
    Sine of an angle with the angle given in degrees.

    Args:
        angle (float): Angle in degrees.

    Returns:
        float
    """
    return np.sin(np.radians(angle))


def tand(angle):
    """
    Tangent of an angle with the angle given in degrees.

    Args:
        angle (float): Angle in degrees.

    Returns:
        float
    """
    return np.tan(np.radians(angle))


def wrap_180(x):
    """
    Shift the given values to within the range (-180, 180].

    Args:
        x (numeric or np.array): Scalar value or np.array of values to shift.

    Returns:
        np.array: Shifted values.
    """
    x = np.where(x <= -180.0, x + 360.0, x)
    x = np.where(x > 180.0, x - 360.0, x)
    return x


def wrap_360(x):
    """
    Shift the given values to within the range (0, 360].

    Args:
        x (numeric or np.array): Scalar value or np.array of values to shift.

    Returns:
        np.array: Shifted values.
    """
    x = np.where(x < 0.0, x + 360.0, x)
    x = np.where(x >= 360.0, x - 360.0, x)
    return x


def wind_delta(wind_directions):
    """
    This is included as a function in order to facilitate testing.
    """
    return ((wind_directions - 270) % 360 + 360) % 360


def rotate_coordinates_rel_west(wind_directions, coordinates):
    # Calculate the difference in given wind direction from 270 / West
    wind_deviation_from_west = wind_delta(wind_directions)
    wind_deviation_from_west = np.reshape(wind_deviation_from_west, (len(wind_directions), 1, 1))

    # Construct the arrays storing the turbine locations
    x_coordinates, y_coordinates, z_coordinates = coordinates.T

    # Find center of rotation - this is the center of box bounding all of the turbines
    x_center_of_rotation = (np.min(x_coordinates) + np.max(x_coordinates)) / 2
    y_center_of_rotation = (np.min(y_coordinates) + np.max(y_coordinates)) / 2

    # Rotate turbine coordinates about the center
    x_coord_offset = x_coordinates - x_center_of_rotation
    y_coord_offset = y_coordinates - y_center_of_rotation
    x_coord_rotated = (
        x_coord_offset * cosd(wind_deviation_from_west)
        - y_coord_offset * sind(wind_deviation_from_west)
        + x_center_of_rotation
    )
    y_coord_rotated = (
        x_coord_offset * sind(wind_deviation_from_west)
        + y_coord_offset * cosd(wind_deviation_from_west)
        + y_center_of_rotation
    )
    z_coord_rotated = np.ones_like(wind_deviation_from_west) * z_coordinates
    return x_coord_rotated, y_coord_rotated, z_coord_rotated



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np


class BaseCOE:
    """
    BaseCOE is the base cost of energy (COE) class that is used to determine
    the cost of energy associated with a
    :py:class:`~.optimization.scipy.layout_height.LayoutHeightOptimization`
    object.

    TODO: 1) Add references to NREL 2016 Cost of Wind Energy Review throughout?
    """

    def __init__(self, opt_obj):
        """
        Instantiate a COE model object with a LayoutHeightOptimization object.

        Args:
            opt_obj (:py:class:`~.layout_height.LayoutHeightOptimization`):
            The optimization object.
        """
        self.opt_obj = opt_obj

    # Public methods

    def FCR(self):
        """
        This method returns the fixed charge rate used in the COE calculation.

        Returns:
            float: The fixed charge rate.
        """
        return 0.079  # % - Taken from 2016 Cost of Wind Energy Review

    def TCC(self, height):
        """
        This method dertermines the turbine capital costs (TCC),
        calculating the effect of varying turbine height and rotor
        diameter on the cost of the tower. The relationship estiamted
        the mass of steel needed for the tower from the NREL Cost and
        Scaling Model (CSM), and then adds that to the tower cost
        portion of the TCC. The proportion is determined from the NREL
        2016 Cost of Wind Energy Review. A price of 3.08 $/kg is
        assumed for the needed steel. Tower height is passed directly
        while the turbine rotor diameter is pulled directly from the
        turbine object within the
        :py:class:`~.tools.floris_interface.FlorisInterface`:.

        TODO: Turbine capital cost or tower capital cost?

        Args:
            height (float): Turbine hub height in meters.

        Returns:
            float: The turbine capital cost of a wind plant in units of $/kWh.
        """
        # From CSM with a fudge factor
        tower_mass = (
            0.2694
            * height
            * (
                np.pi
                * (self.opt_obj.fi.floris.farm.turbines[0].rotor_diameter / 2) ** 2
            )
            + 1779.3
        ) / (1.341638)

        # Combo of 2016 Cost of Wind Energy Review and CSM
        TCC = 831 + tower_mass * 3.08 * self.opt_obj.nturbs / self.opt_obj.plant_kw

        return TCC

    def BOS(self):
        """
        This method returns the balance of station cost of a wind plant as
        determined by a constant factor. As the rating of a wind plant grows,
        the cost of the wind plant grows as well.

        Returns:
            float: The balance of station cost of a wind plant in units of
            $/kWh.
        """
        return 364.0  # $/kW - Taken from 2016 Cost of Wind Energy Review

    def FC(self):
        """
        This method returns the finance charge cost of a wind plant as
        determined by a constant factor. As the rating of a wind plant grows,
        the cost of the wind plant grows as well.

        Returns:
            float: The finance charge cost of a wind plant in units of $/kWh.
        """
        return 155.0  # $/kW - Taken from 2016 Cost of Wind Energy Review

    def O_M(self):
        """
        This method returns the operational cost of a wind plant as determined
        by a constant factor. As the rating of a wind plant grows, the cost of
        the wind plant grows as well.

        Returns:
            float: The operational cost of a wind plant in units of $/kWh.
        """
        return 52.0  # $/kW - Taken from 2016 Cost of Wind Energy Review

    def COE(self, height, AEP_sum):
        """
        This method calculates and returns the cost of energy of a wind plant.
        This cost of energy (COE) formulation for a wind plant varies based on
        turbine height, rotor diameter, and total annualized energy production
        (AEP). The components of the COE equation are defined throughout the
        BaseCOE class.

        Args:
            height (float): The hub height of the turbines in meters
                (all turbines are set to the same height).
            AEP_sum (float): The annualized energy production (AEP)
                for the wind plant as calculated across the wind rose
                in kWh.

        Returns:
            float: The cost of energy for a wind plant in units of
            $/kWh.
        """
        # Comptue Cost of Energy (COE) as $/kWh for a plant
        return (
            self.FCR() * (self.TCC(height) + self.BOS() + self.FC()) + self.O_M()
        ) / (AEP_sum / 1000 / self.opt_obj.plant_kw)



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import copy

import numpy as np
import pandas as pd

from .yaw_wind_rose import YawOptimizationWindRose
from .cluster_turbines import cluster_turbines
from ....logging_manager import LoggerBase


class YawOptimizationWindRoseClustered(YawOptimizationWindRose, LoggerBase):
    """
    YawOptimizationWindRose is a subclass of
    :py:class:`~.tools.optimizationscipy.YawOptimizationWindRose` that is used
    to perform optimizations of the yaw angles of all or a subset of wind
    turbines in a Floris Farm for multiple sets of inflow conditions using the
    scipy optimization package. This class facilitates the clusterization of the
    turbines inside seperate subsets in which the turbines witin each subset
    exclusively interact with one another and have no impact on turbines
    in other clusters. This may significantly reduce the computational
    burden at no loss in performance (assuming the turbine clusters are truly
    independent).
    """

    def __init__(
        self,
        fi,
        wd,
        ws,
        ti=None,
        minimum_yaw_angle=0.0,
        maximum_yaw_angle=25.0,
        minimum_ws=3.0,
        maximum_ws=25.0,
        yaw_angles_baseline=None,
        x0=None,
        bnds=None,
        opt_method="SLSQP",
        opt_options=None,
        include_unc=False,
        unc_pmfs=None,
        unc_options=None,
        turbine_weights=None,
        verbose=False,
        calc_init_power=True,
        exclude_downstream_turbines=False,
        clustering_wake_slope=0.30,
    ):
        """
        Instantiate YawOptimizationWindRose object with a FlorisInterface object
        and assign parameter values.

        Args:
            fi (:py:class:`~.tools.floris_interface.FlorisInterface`):
                Interface used to interact with the Floris object.
            wd (iterable) : The wind directions for which the yaw angles are
                optimized (deg).
            ws (iterable): The wind speeds for which the yaw angles are
                optimized (m/s).
            ti (iterable, optional): An optional list of turbulence intensity
                values for which the yaw angles are optimized. If not
                specified, the current TI value in the Floris object will be
                used for all optimizations. Defaults to None.
            minimum_yaw_angle (float, optional): Minimum constraint on yaw
                angle (deg). This value will be ignored if bnds is also
                specified. Defaults to 0.0.
            maximum_yaw_angle (float, optional): Maximum constraint on yaw
                angle (deg). This value will be ignored if bnds is also
                specified. Defaults to 25.0.
            minimum_ws (float, optional): Minimum wind speed at which
                optimization is performed (m/s). Assumes zero power generated
                below this value. Defaults to 3.
            maximum_ws (float, optional): Maximum wind speed at which
                optimization is performed (m/s). Assumes optimal yaw offsets
                are zero above this wind speed. Defaults to 25.
            yaw_angles_baseline (iterable, optional): The baseline yaw
                angles used to calculate the initial and baseline power
                production in the wind farm and used to normalize the cost
                function. If none are specified, this variable is set equal
                to the current yaw angles in floris. Note that this variable
                need not meet the yaw constraints specified in self.bnds,
                yet a warning is raised if it does to inform the user.
                Defaults to None.
            x0 (iterable, optional): The initial guess for the optimization
                problem. These values must meet the constraints specified
                in self.bnds. Note that, if exclude_downstream_turbines=True,
                the initial guess for any downstream turbines are ignored
                since they are not part of the optimization. Instead, the yaw
                angles for those turbines are 0.0 if that meets the lower and
                upper bound, or otherwise as close to 0.0 as feasible. If no
                values for x0 are specified, x0 is set to be equal to zeros
                wherever feasible (w.r.t. the bounds), and equal to the
                average of its lower and upper bound for all non-downstream
                turbines otherwise. Defaults to None.
            bnds (iterable, optional): Bounds for the yaw angles, as tuples of
                min, max values for each turbine (deg). One can fix the yaw
                angle of certain turbines to a predefined value by setting that
                turbine's lower bound equal to its upper bound (i.e., an
                equality constraint), as: bnds[ti] = (x, x), where x is the
                fixed yaw angle assigned to the turbine. This works for both
                zero and nonzero yaw angles. Moreover, if 
                exclude_downstream_turbines=True, the yaw angles for all
                downstream turbines will be 0.0 or a feasible value closest to
                0.0. If none are specified, the bounds are set to
                (minimum_yaw_angle, maximum_yaw_angle) for each turbine. Note
                that, if bnds is not none, its values overwrite any value given
                in minimum_yaw_angle and maximum_yaw_angle. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to 'SLSQP'.
            opt_options (dictionary, optional): Optimization options used by
                scipy.optimize.minize. If none are specified, they are set to
                {'maxiter': 100, 'disp': False, 'iprint': 1, 'ftol': 1e-7,
                'eps': 0.01}. Defaults to None.
            include_unc (bool, optional): Determines whether wind direction or
                yaw uncertainty are included. If True, uncertainty in wind
                direction and/or yaw position is included when determining
                wind farm power. Uncertainty is included by computing the
                mean wind farm power for a distribution of wind direction
                and yaw position deviations from the intended wind direction
                and yaw angles. Defaults to False.
            unc_pmfs (dictionary, optional): A dictionary containing
                probability mass functions describing the distribution of
                wind direction and yaw position deviations when wind direction
                and/or yaw position uncertainty is included in the power
                calculations. Contains the following key-value pairs:

                -   **wd_unc** (*np.array*): The wind direction
                    deviations from the intended wind direction (deg).
                -   **wd_unc_pmf** (*np.array*): The probability
                    of each wind direction deviation in **wd_unc** occuring.
                -   **yaw_unc** (*np.array*): The yaw angle deviations
                    from the intended yaw angles (deg).
                -   **yaw_unc_pmf** (*np.array*): The probability
                    of each yaw angle deviation in **yaw_unc** occuring.

                If none are specified, default PMFs are calculated using
                values provided in **unc_options**. Defaults to None.
            unc_options (dictionary, optional): A dictionary containing values
                used to create normally-distributed, zero-mean probability mass
                functions describing the distribution of wind direction and yaw
                position deviations when wind direction and/or yaw position
                uncertainty is included. This argument is only used when
                **unc_pmfs** is None and contains the following key-value pairs:

                -   **std_wd** (*float*): The standard deviation of
                    the wind direction deviations from the original wind
                    direction (deg).
                -   **std_yaw** (*float*): The standard deviation of
                    the yaw angle deviations from the original yaw angles (deg).
                -   **pmf_res** (*float*): The resolution in degrees
                    of the wind direction and yaw angle PMFs.
                -   **pdf_cutoff** (*float*): The cumulative
                    distribution function value at which the tails of the
                    PMFs are truncated.

                If none are specified, default values of
                {'std_wd': 4.95, 'std_yaw': 1.75, 'pmf_res': 1.0,
                'pdf_cutoff': 0.995} are used. Defaults to None.
            turbine_weights (iterable, optional): weighing terms that allow
                the user to emphasize power gains at particular turbines or
                completely ignore power gains from other turbines. The array
                of turbine powers from floris is multiplied with this array
                in the calculation of the objective function. If None, this
                is an array with all values 1.0 and length equal to the
                number of turbines. Defaults to None.
            calc_init_power (bool, optional): If True, calculates initial
                wind farm power for each set of wind conditions. Defaults to
                True.
            exclude_downstream_turbines (bool, optional): If True,
                automatically finds and excludes turbines that are most
                downstream from the optimization problem. This significantly
                reduces computation time at no loss in performance. The yaw
                angles of these downstream turbines are fixed to 0.0 deg if
                the yaw bounds specified in self.bnds allow that, or otherwise
                are fixed to the lower or upper yaw bound, whichever is closer
                to 0.0. Defaults to False.
            clustering_wake_slope (float, optional): linear slope of the wake
                in the simplified linear expansion wake model (dy/dx). This
                model is used to derive wake interactions between turbines and
                to identify the turbine clusters. A good value is about equal
                to the turbulence intensity in FLORIS. Though, since yaw
                optimizations may shift the wake laterally, a safer option
                is twice the turbulence intensity. The default value is 0.30
                which should be valid for yaw optimizations at wd_std = 0.0 deg
                and turbulence intensities up to 15%. Defaults to 0.30.
        """
        super().__init__(
            fi=fi,
            wd=wd,
            ws=ws,
            ti=ti,
            minimum_yaw_angle=minimum_yaw_angle,
            maximum_yaw_angle=maximum_yaw_angle,
            minimum_ws=minimum_ws,
            maximum_ws=maximum_ws,
            yaw_angles_baseline=yaw_angles_baseline,
            x0=x0,
            bnds=bnds,
            opt_method=opt_method,
            opt_options=opt_options,
            include_unc=include_unc,
            unc_pmfs=unc_pmfs,
            unc_options=unc_options,
            turbine_weights=turbine_weights,
            verbose=verbose,
            calc_init_power=calc_init_power,
            exclude_downstream_turbines=exclude_downstream_turbines,
        )
        self.clustering_wake_slope = clustering_wake_slope


    def _cluster_turbines(self):
        wind_directions = self.fi.floris.farm.wind_direction
        if (np.std(wind_directions) > 0.001):
            raise ValueError("Wind directions must be uniform for clustering algorithm.")
        self.clusters = cluster_turbines(
            fi=self.fi,
            wind_direction=self.fi.floris.farm.wind_direction[0],
            wake_slope=self.clustering_wake_slope
        )

    def plot_clusters(self):
        for wd in self.wd:
            cluster_turbines(
                fi=self.fi,
                wind_direction=wd,
                wake_slope=self.clustering_wake_slope,
                plot_lines=True
            )


    def optimize(self):
        """
        This method solves for the optimum turbine yaw angles for power
        production and the resulting power produced by the wind farm for a
        series of wind speed, wind direction, and optionally TI combinations.

        Returns:
            pandas.DataFrame: A pandas DataFrame with the same number of rows
            as the length of the wd and ws arrays, containing the following
            columns:

                - **ws** (*float*) - The wind speed values for which the yaw
                angles are optimized and power is computed (m/s).
                - **wd** (*float*) - The wind direction values for which the
                yaw angles are optimized and power is computed (deg).
                - **ti** (*float*) - The turbulence intensity values for which
                the yaw angles are optimized and power is computed. Only
                included if self.ti is not None.
                - **power_opt** (*float*) - The total power produced by the
                wind farm with optimal yaw offsets (W).
                - **turbine_power_opt** (*list* (*float*)) - A list
                containing the power produced by each wind turbine with optimal
                yaw offsets (W).
                - **yaw_angles** (*list* (*float*)) - A list containing
                the optimal yaw offsets for maximizing total wind farm power
                for each wind turbine (deg).
        """
        print("=====================================================")
        print("Optimizing wake redirection control...")
        print("Number of wind conditions to optimize = ", len(self.wd))
        print("Number of yaw angles to optimize = ", len(self.turbs_to_opt))
        print("=====================================================")

        df_opt = pd.DataFrame()

        for i in range(len(self.wd)):
            if self.verbose:
                if self.ti is None:
                    print(
                        "Computing wind speed, wind direction pair "
                        + str(i)
                        + " out of "
                        + str(len(self.wd))
                        + ": wind speed = "
                        + str(self.ws[i])
                        + " m/s, wind direction = "
                        + str(self.wd[i])
                        + " deg."
                    )
                else:
                    print(
                        "Computing wind speed, wind direction, turbulence "
                        + "intensity set "
                        + str(i)
                        + " out of "
                        + str(len(self.wd))
                        + ": wind speed = "
                        + str(self.ws[i])
                        + " m/s, wind direction = "
                        + str(self.wd[i])
                        + " deg, turbulence intensity = "
                        + str(self.ti[i])
                        + "."
                    )

            # Optimizing wake redirection control
            if (self.ws[i] >= self.minimum_ws) & (self.ws[i] <= self.maximum_ws):
                if self.ti is None:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]], wind_speed=[self.ws[i]]
                    )
                else:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]],
                        wind_speed=[self.ws[i]],
                        turbulence_intensity=self.ti[i],
                    )

                # Set initial farm power
                self.initial_farm_power = self.initial_farm_powers[i]

                # Determine clusters and then optimize by cluster
                self._cluster_turbines()
                if self.verbose:
                    print("Clustered turbines into %d separate clusters." % len(self.clusters))

                # Save parameters to a full list
                yaw_angles_template_full = copy.copy(self.yaw_angles_template)
                yaw_angles_baseline_full = copy.copy(self.yaw_angles_baseline)
                turbine_weights_full = copy.copy(self.turbine_weights)
                bnds_full = copy.copy(self.bnds)
                # nturbs_full = copy.copy(self.nturbs)
                x0_full = copy.copy(self.x0)
                fi_full = copy.deepcopy(self.fi)

                # Overwrite parameters for each cluster and optimize
                opt_yaw_angles = np.zeros_like(x0_full)
                for ci, cl in enumerate(self.clusters):
                    if self.verbose:
                        print("=====================================================")
                        print("Optimizing %d parameters in cluster %d." % (len(cl), ci))
                        print("=====================================================")
                    self.yaw_angles_template = np.array(yaw_angles_template_full)[cl]
                    self.yaw_angles_baseline = np.array(yaw_angles_baseline_full)[cl]
                    self.turbine_weights = np.array(turbine_weights_full)[cl]
                    self.bnds = np.array(bnds_full)[cl]
                    self.x0 = np.array(x0_full)[cl]
                    self.fi = copy.deepcopy(fi_full)
                    self.fi.reinitialize_flow_field(
                        layout_array=[
                            np.array(fi_full.layout_x)[cl],
                            np.array(fi_full.layout_y)[cl]
                        ]
                    )
                    opt_yaw_angles[cl] = self._optimize()

                # Restore parameters
                self.yaw_angles_template = yaw_angles_template_full
                self.yaw_angles_baseline = yaw_angles_baseline_full
                self.turbine_weights = turbine_weights_full
                self.bnds = bnds_full
                self.x0 = x0_full
                self.fi = fi_full
                self.fi.reinitialize_flow_field(
                    layout_array=[
                        np.array(fi_full.layout_x),
                        np.array(fi_full.layout_y)
                    ]
                )

                if np.sum(np.abs(opt_yaw_angles)) == 0:
                    print(
                        "No change in controls suggested for this inflow \
                        condition..."
                    )

                # optimized power
                self.fi.calculate_wake(yaw_angles=opt_yaw_angles)
                power_opt = self.fi.get_turbine_power(
                    include_unc=self.include_unc,
                    unc_pmfs=self.unc_pmfs,
                    unc_options=self.unc_options,
                )
            elif self.ws[i] >= self.maximum_ws:
                print(
                    "No change in controls suggested for this inflow \
                        condition..."
                )
                if self.ti is None:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]], wind_speed=[self.ws[i]]
                    )
                else:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]],
                        wind_speed=[self.ws[i]],
                        turbulence_intensity=self.ti[i],
                    )
                opt_yaw_angles = np.array(self.yaw_angles_template, copy=True)
                self.fi.calculate_wake(yaw_angles=opt_yaw_angles)
                power_opt = self.fi.get_turbine_power(
                    include_unc=self.include_unc,
                    unc_pmfs=self.unc_pmfs,
                    unc_options=self.unc_options,
                )
            else:
                print(
                    "No change in controls suggested for this inflow \
                        condition..."
                )
                opt_yaw_angles = np.array(self.yaw_angles_template, copy=True)
                power_opt = self.nturbs * [0.0]

            # Include turbine weighing terms
            power_opt = np.multiply(self.turbine_weights, power_opt)

            # add variables to dataframe
            if self.ti is None:
                df_opt = df_opt.append(
                    pd.DataFrame(
                        {
                            "ws": [self.ws[i]],
                            "wd": [self.wd[i]],
                            "power_opt": [np.sum(power_opt)],
                            "turbine_power_opt": [power_opt],
                            "yaw_angles": [opt_yaw_angles],
                        }
                    )
                )
            else:
                df_opt = df_opt.append(
                    pd.DataFrame(
                        {
                            "ws": [self.ws[i]],
                            "wd": [self.wd[i]],
                            "ti": [self.ti[i]],
                            "power_opt": [np.sum(power_opt)],
                            "turbine_power_opt": [power_opt],
                            "yaw_angles": [opt_yaw_angles],
                        }
                    )
                )

        df_opt.reset_index(drop=True, inplace=True)

        return df_opt



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import inspect


def show_params(
    fi,
    params=None,
    verbose=False,
    wake_velocity_model=True,
    wake_deflection_model=True,
    turbulence_model=True,
):

    if wake_velocity_model:
        obj = "fi.floris.wake.velocity_model"
        # props = get_props(obj, fi)
        props = fi.floris.wake._asdict()
        # props = props["wake_velocity_parameters"][fi.floris.wake.velocity_model.model_string]
        props = fi.floris.wake.velocity_model._get_model_dict()

        if verbose:
            print("=".join(["="] * 39))
        else:
            print("=".join(["="] * 19))
        print(
            "Wake Velocity Model Parameters:",
            fi.floris.wake.velocity_model.model_string,
            "model",
        )

        if params is not None:
            props_subset = get_props_subset(params, props)
            if not verbose:
                print_props(obj, fi, props_subset)
            else:
                print_prop_docs(obj, fi, props_subset)

        else:
            if not verbose:
                print_props(obj, fi, props)
            else:
                print_prop_docs(obj, fi, props)

    if wake_deflection_model:
        obj = "fi.floris.wake.deflection_model"
        props = get_props(obj, fi)

        if verbose:
            print("=".join(["="] * 39))
        else:
            print("=".join(["="] * 19))
        print(
            "Wake Deflection Model Parameters:",
            fi.floris.wake.deflection_model.model_string,
            "model",
        )

        if params is not None:
            props_subset = get_props_subset(params, props)
            if props_subset:  # true if the subset is not empty
                if not verbose:
                    print_props(obj, fi, props_subset)
                else:
                    print_prop_docs(obj, fi, props_subset)

        else:
            if not verbose:
                print_props(obj, fi, props)
            else:
                print_prop_docs(obj, fi, props)

    if turbulence_model:
        obj = "fi.floris.wake.turbulence_model"
        props = get_props(obj, fi)

        if verbose:
            print("=".join(["="] * 39))
        else:
            print("=".join(["="] * 19))
        print(
            "Wake Turbulence Model Parameters:",
            fi.floris.wake.turbulence_model.model_string,
            "model",
        )

        if params is not None:
            props_subset = get_props_subset(params, props)
            if props_subset:  # true if the subset is not empty
                if not verbose:
                    print_props(obj, fi, props_subset)
                else:
                    print_prop_docs(obj, fi, props_subset)

        else:
            if not verbose:
                print_props(obj, fi, props)
            else:
                print_prop_docs(obj, fi, props)


def get_params(
    fi,
    params=None,
    wake_velocity_model=True,
    wake_deflection_model=True,
    turbulence_model=True,
):
    model_params = {}

    if wake_velocity_model:
        wake_vel_vals = {}
        obj = "fi.floris.farm.wake.velocity_model"
        props = get_props(obj, fi)
        if params is not None:
            props_subset = get_props_subset(params, props)
            wake_vel_vals = get_prop_values(obj, fi, props_subset)
        else:
            wake_vel_vals = get_prop_values(obj, fi, props)
        model_params["Wake Velocity Parameters"] = wake_vel_vals
        del model_params["Wake Velocity Parameters"]["logger"]

    if wake_deflection_model:
        wake_defl_vals = {}
        obj = "fi.floris.farm.wake.deflection_model"
        props = get_props(obj, fi)
        if params is not None:
            props_subset = get_props_subset(params, props)
            wake_defl_vals = get_prop_values(obj, fi, props_subset)
        else:
            wake_defl_vals = get_prop_values(obj, fi, props)
        model_params["Wake Deflection Parameters"] = wake_defl_vals
        del model_params["Wake Deflection Parameters"]["logger"]

    if turbulence_model:
        wake_turb_vals = {}
        obj = "fi.floris.farm.wake.turbulence_model"
        props = get_props(obj, fi)
        if params is not None:
            props_subset = get_props_subset(params, props)
            wake_turb_vals = get_prop_values(obj, fi, props_subset)
        else:
            wake_turb_vals = get_prop_values(obj, fi, props)
        model_params["Wake Turbulence Parameters"] = wake_turb_vals
        del model_params["Wake Turbulence Parameters"]["logger"]

    return model_params


def set_params(fi, params, verbose=True):
    for param_dict in params:
        if param_dict == "Wake Velocity Parameters":
            obj = "fi.floris.farm.wake.velocity_model"
            props = get_props(obj, fi)
            for prop in params[param_dict]:
                if prop in [val[0] for val in props]:
                    exec(obj + "." + prop + " = " + str(params[param_dict][prop]))
                    if verbose:
                        print(
                            "Wake velocity parameter "
                            + prop
                            + " set to "
                            + str(params[param_dict][prop])
                        )
                else:
                    raise Exception(
                        (
                            "Wake deflection parameter '{}' "
                            + "not part of current model. Value '{}' was not "
                            + "used."
                        ).format(prop, params[param_dict][prop])
                    )

        if param_dict == "Wake Deflection Parameters":
            obj = "fi.floris.farm.wake.deflection_model"
            props = get_props(obj, fi)
            for prop in params[param_dict]:
                if prop in [val[0] for val in props]:
                    exec(obj + "." + prop + " = " + str(params[param_dict][prop]))
                    if verbose:
                        print(
                            "Wake deflection parameter "
                            + prop
                            + " set to "
                            + str(params[param_dict][prop])
                        )
                else:
                    raise Exception(
                        (
                            "Wake deflection parameter '{}' "
                            + "not part of current model. Value '{}' was not "
                            + "used."
                        ).format(prop, params[param_dict][prop])
                    )

        if param_dict == "Wake Turbulence Parameters":
            obj = "fi.floris.farm.wake.turbulence_model"
            props = get_props(obj, fi)
            for prop in params[param_dict]:
                if prop in [val[0] for val in props]:
                    exec(obj + "." + prop + " = " + str(params[param_dict][prop]))
                    if verbose:
                        print(
                            "Wake turbulence parameter "
                            + prop
                            + " set to "
                            + str(params[param_dict][prop])
                        )
                else:
                    raise Exception(
                        (
                            "Wake turbulence parameter '{}' "
                            + "not part of current model. Value '{}' was not "
                            + "used."
                        ).format(prop, params[param_dict][prop])
                    )


def get_props_subset(params, props):
    prop_names = [prop[0] for prop in props]
    try:
        props_subset_inds = [prop_names.index(param) for param in params]
    except Exception:
        props_subset_inds = []
        print("Parameter(s)", ", ".join(params), "does(do) not exist.")
    props_subset = [props[i] for i in props_subset_inds]
    return props_subset


# def get_props(obj, fi):
#     return inspect.getmembers(
#         eval(obj + ".__class__"), lambda obj: isinstance(obj, property)
#     )


def get_prop_values(obj, fi, props):
    prop_val_dict = {}
    for val in props:
        prop_val_dict[val[0]] = eval(obj + "." + val[0])
    return prop_val_dict


def print_props(obj, fi, props):
    print("-".join(["-"] * 19))
    for val in props:
        print(val[0] + " = " + str(eval(obj + "." + val[0])))
    print("-".join(["-"] * 19))


def print_prop_docs(obj, fi, props):
    for val in props:
        print(
            "-".join(["-"] * 39) + "\n",
            val[0] + " = " + str(eval(obj + "." + val[0])),
            "\n",
            eval(obj + ".__class__." + val[0] + ".__doc__"),
        )
    print("-".join(["-"] * 39))



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable

import attrs
from attrs import define, field
import numpy as np

from floris.utilities import Vec3, rotate_coordinates_rel_west, cosd, sind
from floris.type_dec import  (
    floris_float_type,
    floris_array_converter,
    NDArrayFloat,
    NDArrayInt
)

@define
class Grid(ABC):
    """
    Grid should establish domain bounds based on given criteria,
    and develop three arrays to contain components of the grid
    locations in space.

    This could be generalized to any number of dimensions to be
    used by perhaps a turbulence field.

    The grid will have to be reestablished for each wind direction since the planform
    area of the farm will be different.

    x are the locations in space in the primary direction (typically the direction of the wind)
    y are the locations in space in the lateral direction
    z are the locations in space in the vertical direction
    u are the velocity components at each point in space
    v are the velocity components at each point in space
    w are the velocity components at each point in space
    all of these arrays are the same size

    Args:
        turbine_coordinates (`list[Vec3]`): The collection of turbine coordinate (`Vec3`) objects.
        reference_turbine_diameter (:py:obj:`float`): The reference turbine's rotor diameter.
        grid_resolution (:py:obj:`int` | :py:obj:`Iterable(int,)`): Grid resolution specific to each grid type
    """
    turbine_coordinates: list[Vec3] = field()
    reference_turbine_diameter: float
    grid_resolution: int | Iterable = field()
    wind_directions: NDArrayFloat = field(converter=floris_array_converter)
    wind_speeds: NDArrayFloat = field(converter=floris_array_converter)

    n_turbines: int = field(init=False)
    n_wind_speeds: int = field(init=False)
    n_wind_directions: int = field(init=False)
    turbine_coordinates_array: NDArrayFloat = field(init=False)
    x: NDArrayFloat = field(init=False)
    y: NDArrayFloat = field(init=False)
    z: NDArrayFloat = field(init=False)

    def __attrs_post_init__(self) -> None:
        self.turbine_coordinates_array = np.array([c.elements for c in self.turbine_coordinates])

    @turbine_coordinates.validator
    def check_coordinates(self, instance: attrs.Attribute, value: list[Vec3]) -> None:
        """Ensures all elements are `Vec3` objects and keeps the `n_turbines` attribute up to date."""
        types = np.unique([isinstance(c, Vec3) for c in value])
        if not all(types):
            raise TypeError("'turbine_coordinates' must be `Vec3` objects.")

        self.n_turbines = len(value)

    @wind_speeds.validator
    def wind_speeds_validator(self, instance: attrs.Attribute, value: NDArrayFloat) -> None:
        """Using the validator method to keep the `n_wind_speeds` attribute up to date."""
        self.n_wind_speeds = value.size

    @wind_directions.validator
    def wind_directions_validator(self, instance: attrs.Attribute, value: NDArrayFloat) -> None:
        """Using the validator method to keep the `n_wind_directions` attribute up to date."""
        self.n_wind_directions = value.size

    @grid_resolution.validator
    def grid_resolution_validator(self, instance: attrs.Attribute, value: int | Iterable) -> None:
        # TODO move this to the grid types and off of the base class
        """Check that grid resolution is given as int or Vec3 with int components."""
        if isinstance(value, int) and type(self) is TurbineGrid:
            return
        elif isinstance(value, Iterable) and type(self) is FlowFieldPlanarGrid:
            assert type(value[0]) is int
            assert type(value[1]) is int
        elif isinstance(value, Iterable) and type(self) is FlowFieldGrid:
            assert type(value[0]) is int
            assert type(value[1]) is int
            assert type(value[2]) is int
        else:
            raise TypeError("`grid_resolution` must be of type int or Iterable(int,)")

    @abstractmethod
    def set_grid(self) -> None:
        raise NotImplementedError("Grid.set_grid")

    @abstractmethod
    def finalize(self) -> None:
        raise NotImplementedError("Grid.finalize")


@define
class TurbineGrid(Grid):
    """See `Grid` for more details.

    Args:
        turbine_coordinates (`list[Vec3]`): The collection of turbine coordinate (`Vec3`) objects.
        reference_turbine_diameter (:py:obj:`float`): The reference turbine's rotor diameter.
        wind_directions (`list[float]`): The input wind directions
        wind_speeds (`list[float]`): The input wind speeds
        grid_resolution (:py:obj:`int`): The number of points on each turbine
    """
    sorted_indices: NDArrayInt = field(init=False)
    unsorted_indices: NDArrayInt = field(init=False)

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self.set_grid()

    def set_grid(self) -> None:
        """
        Create grid points at each turbine for each wind direction and wind speed in the simulation.
        This creates the underlying data structure for the calculation.

        arrays have shape (n wind directions, n wind speeds, n turbines, m grid spanwise, m grid vertically)
        - dimension 1: each wind direction
        - dimension 2: each wind speed
        - dimension 3: each turbine
        - dimension 4: number of points in the spanwise direction (ngrid)
        - dimension 5: number of points in the vertical dimension (ngrid)

        For example
        - x is [n wind direction, n wind speeds, n turbines, x-component of the points in the spanwise direction, x-component of the points in the vertical direction]
        - y is [n wind direction, n wind speeds, n turbines, y-component of the points in the spanwise direction, y-component of the points in the vertical direction]

        The x,y,z arrays contain the actual locations in that direction.

        # -   **self.grid_resolution** (*int*, optional): The square root of the number
        #             of points to use on the turbine grid. This number will be
        #             squared so that the points can be evenly distributed.
        #             Defaults to 5.

        If the grid conforms to the sequential solver interface, it must be sorted from upstream to downstream

        In a y-z plane on the rotor swept area, the -2 dimension is a column of points and the -1 dimension is the row number.
        So the following line prints the 0'th column of the the 0'th turbine's grid:
        print(grid.y[0,0,0,0,:])
        print(grid.z[0,0,0,0,:])
        And this line prints a single point
        print(grid.y[0,0,0,0,0])
        print(grid.z[0,0,0,0,0])
        Note that the x coordinates are all the same for the rotor plane.

        """
        # TODO: Where should we locate the coordinate system? Currently, its at
        # the foot of the turbine where the tower meets the ground.

        # These are the rotated coordinates of the wind turbines based on the wind direction
        x, y, z = rotate_coordinates_rel_west(self.wind_directions, self.turbine_coordinates_array)

        # -   **rloc** (*float, optional): A value, from 0 to 1, that determines
        #         the width/height of the grid of points on the rotor as a ratio of
        #         the rotor radius.
        #         Defaults to 0.5.

        # Create the data for the turbine grids
        radius_ratio = 0.5
        disc_area_radius = radius_ratio * self.reference_turbine_diameter / 2
        template_grid = np.ones(
            (
                self.n_wind_directions,
                self.n_wind_speeds,
                self.n_turbines,
                self.grid_resolution,
                self.grid_resolution,
            ),
            dtype=floris_float_type
        )

        # Calculate the radial distance from the center of the turbine rotor
        disc_grid = np.linspace(-1 * disc_area_radius, disc_area_radius, self.grid_resolution, dtype=floris_float_type)

        # Construct the turbine grids
        # Here, they are already rotated to the correct orientation for each wind direction
        _x = x[:, :, :, None, None] * template_grid

        ones_grid = np.ones((self.grid_resolution, self.grid_resolution), dtype=floris_float_type)
        # [:, None] on disc_grid below is effectively a transpose of this vector; compare with disc_grid.reshape(1,-1).T
        _y = y[:, :, :, None, None] + template_grid * ( disc_grid[:, None] * ones_grid )
        _z = z[:, :, :, None, None] + template_grid * ( disc_grid * ones_grid )

        # Sort the turbines at each wind direction

        # Get the sorted indices for the x coordinates. These are the indices
        # to sort the turbines from upstream to downstream for all wind directions.
        # Also, store the indices to sort them back for when the calculation finishes.
        self.sorted_indices = _x.argsort(axis=2)
        self.unsorted_indices = self.sorted_indices.argsort(axis=2)

        # Put the turbines into the final arrays in their sorted order
        self.x = np.take_along_axis(_x, self.sorted_indices, axis=2)
        self.y = np.take_along_axis(_y, self.sorted_indices, axis=2)
        self.z = np.take_along_axis(_z, self.sorted_indices, axis=2)

    def finalize(self):
        # Replace the turbines in their unsorted order so that
        # we can report values in a sane way.
        self.x = np.take_along_axis(self.x, self.unsorted_indices, axis=2)
        self.y = np.take_along_axis(self.y, self.unsorted_indices, axis=2)
        self.z = np.take_along_axis(self.z, self.unsorted_indices, axis=2)


@define
class FlowFieldGrid(Grid):
    """
    Args:
        grid_resolution (`Vec3`): The number of grid points to be created in each direction.
        turbine_coordinates (`list[Vec3]`): The collection of turbine coordinate (`Vec3`) objects.
        reference_turbine_diameter (:py:obj:`float`): The reference turbine's rotor diameter.
        grid_resolution (:py:obj:`int`): The number of points on each turbine
    """

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self.set_grid()

    def set_grid(self) -> None:
        """
        Create a structured grid for the entire flow field domain.
        resolution: Vec3

        Calculates the domain bounds for the current wake model. The bounds
        are calculated based on preset extents from the
        given layout. The bounds consist of the minimum and maximum values
        in the x-, y-, and z-directions.

        If the Curl model is used, the predefined bounds are always set.

        First, sort the turbines so that we know the bounds in the correct orientation.
        Then, create the grid based on this wind-from-left orientation
        """

        # These are the rotated coordinates of the wind turbines based on the wind direction
        x, y, z = rotate_coordinates_rel_west(self.wind_directions, self.turbine_coordinates_array)

        # Construct the arrays storing the grid points
        eps = 0.01
        xmin = min(x[0,0]) - 2 * self.reference_turbine_diameter
        xmax = max(x[0,0]) + 10 * self.reference_turbine_diameter
        ymin = min(y[0,0]) - 2 * self.reference_turbine_diameter
        ymax = max(y[0,0]) + 2 * self.reference_turbine_diameter
        zmin = 0 + eps
        zmax = 6 * max(z[0,0])

        x_points, y_points, z_points = np.meshgrid(
            np.linspace(xmin, xmax, int(self.grid_resolution[0])),
            np.linspace(ymin, ymax, int(self.grid_resolution[1])),
            np.linspace(zmin, zmax, int(self.grid_resolution[2])),
            indexing="ij"
        )

        self.x = x_points[None, None, :, :, :]
        self.y = y_points[None, None, :, :, :]
        self.z = z_points[None, None, :, :, :]

    def finalize(self):
        pass


@define
class FlowFieldPlanarGrid(Grid):
    """
    Args:
        grid_resolution (`Vec3`): The number of grid points to be created in each direction.
        turbine_coordinates (`list[Vec3]`): The collection of turbine coordinate (`Vec3`) objects.
        reference_turbine_diameter (:py:obj:`float`): The reference turbine's rotor diameter.
        grid_resolution (:py:obj:`int`): The number of points on each turbine
    """
    normal_vector: str = field()
    planar_coordinate: float = field()
    x1_bounds: tuple = field(default=None)
    x2_bounds: tuple = field(default=None)

    sorted_indices: NDArrayInt = field(init=False)
    unsorted_indices: NDArrayInt = field(init=False)

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self.set_grid()

    def set_grid(self) -> None:
        """
        Create a structured grid for the entire flow field domain.
        resolution: Vec3

        Calculates the domain bounds for the current wake model. The bounds
        are calculated based on preset extents from the
        given layout. The bounds consist of the minimum and maximum values
        in the x-, y-, and z-directions.

        If the Curl model is used, the predefined bounds are always set.

        First, sort the turbines so that we know the bounds in the correct orientation.
        Then, create the grid based on this wind-from-left orientation
        """
        # These are the rotated coordinates of the wind turbines based on the wind direction
        x, y, z = rotate_coordinates_rel_west(self.wind_directions, self.turbine_coordinates_array)

        max_diameter = self.reference_turbine_diameter

        if self.normal_vector == "z":  # Rules of thumb for horizontal plane
            if self.x1_bounds is None:
                self.x1_bounds = (np.min(x) - 2 * max_diameter, np.max(x) + 10 * max_diameter)

            if self.x2_bounds is None:
                self.x2_bounds = (np.min(y) - 2 * max_diameter, np.max(y) + 2 * max_diameter)

            # TODO figure out proper z spacing for GCH, currently set to +/- 10.0
            x_points, y_points, z_points = np.meshgrid(
                np.linspace(self.x1_bounds[0], self.x1_bounds[1], int(self.grid_resolution[0])),
                np.linspace(self.x2_bounds[0], self.x2_bounds[1], int(self.grid_resolution[1])),
                np.array([float(self.planar_coordinate) - 10.0, float(self.planar_coordinate), float(self.planar_coordinate) + 10.0]),
                indexing="ij"
            )

            self.x = x_points[None, None, :, :, :]
            self.y = y_points[None, None, :, :, :]
            self.z = z_points[None, None, :, :, :]

        elif self.normal_vector == "x":  # Rules of thumb for cross plane
            if self.x1_bounds is None:
                self.x1_bounds = (np.min(y) - 2 * max_diameter, np.max(y) + 2 * max_diameter)

            if self.x2_bounds is None:
                self.x2_bounds = (0.001, 6 * np.max(z))

            x_points, y_points, z_points = np.meshgrid(
                np.array([float(self.planar_coordinate)]),
                np.linspace(self.x1_bounds[0], self.x1_bounds[1], int(self.grid_resolution[0])),
                np.linspace(self.x2_bounds[0], self.x2_bounds[1], int(self.grid_resolution[1])),
                indexing="ij"
            )

            self.x = x_points[None, None, :, :, :]
            self.y = y_points[None, None, :, :, :]
            self.z = z_points[None, None, :, :, :]

        elif self.normal_vector == "y":  # Rules of thumb for y plane
            if self.x1_bounds is None:
                self.x1_bounds = (np.min(x) - 2 * max_diameter, np.max(x) + 10 * max_diameter)

            if self.x2_bounds is None:
                self.x2_bounds = (0.001, 6 * np.max(z))

            x_points, y_points, z_points = np.meshgrid(
                np.linspace(self.x1_bounds[0], self.x1_bounds[1], int(self.grid_resolution[0])),
                np.array([float(self.planar_coordinate)]),
                np.linspace(self.x2_bounds[0], self.x2_bounds[1], int(self.grid_resolution[1])),
                indexing="ij"
            )

            self.x = x_points[None, None, :, :, :]
            self.y = y_points[None, None, :, :, :]
            self.z = z_points[None, None, :, :, :]

        # self.sorted_indices = self.x.argsort(axis=2)
        # self.unsorted_indices = self.sorted_indices.argsort(axis=2)

        # Put the turbines into the final arrays in their sorted order
        # self.x = np.take_along_axis(self.x, self.sorted_indices, axis=2)
        # self.y = np.take_along_axis(self.y, self.sorted_indices, axis=2)
        # self.z = np.take_along_axis(self.z, self.sorted_indices, axis=2)

    def finalize(self):
        # sorted_indices = self.x.argsort(axis=2)
        # unsorted_indices = sorted_indices.argsort(axis=2)

        # # print(self.x)

        # x_coordinates, y_coordinates, _ = self.turbine_coordinates_array.T
        
        # x_center_of_rotation = (np.min(x_coordinates) + np.max(x_coordinates)) / 2
        # y_center_of_rotation = (np.min(y_coordinates) + np.max(y_coordinates)) / 2
        # # print(x_center_of_rotation)
        # # print(y_center_of_rotation)
        # # lkj

        # self.x = np.take_along_axis(self.x, self.unsorted_indices, axis=2)
        # self.y = np.take_along_axis(self.y, self.unsorted_indices, axis=2)
        # self.z = np.take_along_axis(self.z, self.unsorted_indices, axis=2)
        # # print(self.x)

        # self.x, self.y, self.z = self._rotated_grid(-1 * self.wind_directions, (x_center_of_rotation, y_center_of_rotation))
        # TODO figure out how to un-rotate grid for plotting after it has been solved
        pass

    # def _rotated_grid(self, angle, center_of_rotation):
    #     """
    #     Rotate the discrete flow field grid.
    #     """
    #     angle = ((angle - 270) % 360 + 360) % 360
    #     # angle = np.reshape(angle, (len(angle), 1, 1))
    #     xoffset = self.x - center_of_rotation[0]
    #     yoffset = self.y - center_of_rotation[1]
    #     rotated_x = (
    #         xoffset * cosd(angle) - yoffset * sind(angle) + center_of_rotation[0]
    #     )
    #     rotated_y = (
    #         xoffset * sind(angle) + yoffset * cosd(angle) + center_of_rotation[1]
    #     )
    #     return rotated_x, rotated_y, self.z


# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import numpy as np
import matplotlib.pyplot as plt


def derive_downstream_turbines(fi, wind_direction, wake_slope=0.30, plot_lines=False):
    """Determine which turbines have no effect on other turbines in the
    farm, i.e., which turbines have wakes that do not impact the other
    turbines in the farm. This allows the user to exclude these turbines
    from a control setpoint optimization, for example. This function
    assumes a very simplified wake function where the wakes are assumed
    to have a linearly diverging profile. In comparisons with the FLORIS
    GCH model, the wake_slope matches well with the FLORIS' wake profiles
    for a value of wake_slope = 0.5 * turbulence_intensity, where
    turbulence_intensity is an input to the FLORIS model at the default
    GCH parameterization. Note that does not include wind direction variability.
    To be conservative, the user is recommended to use the rule of thumb:
    `wake_slope = turbulence_intensity`. Hence, the default value for
    `wake_slope=0.30` should be conservative for turbulence intensities up to
    0.30 and is likely to provide valid estimates of which turbines are
    downstream until a turbulence intensity of 0.50. This simple model saves
    time compared to FLORIS.

    Args:
        fi ([floris object]): FLORIS object of the farm of interest.
        wind_direction (float): The wind direction in the FLORIS frame
        of reference for which the downstream turbines are to be determined.
        wake_slope (float, optional): linear slope of the wake (dy/dx)
        plot_lines (bool, optional): Enable plotting wakes/turbines.
        Defaults to False.

    Returns:
        turbs_downstream (iterable): A list containing the turbine
        numbers that have a wake that does not affect any other
        turbine inside the farm.
    """

    # Get farm layout
    x = fi.layout_x
    y = fi.layout_y
    D = np.array([t.rotor_diameter for t in fi.floris.farm.turbines])
    n_turbs = len(x)

    # Rotate farm and determine freestream/waked turbines
    is_downstream = [False for _ in range(n_turbs)]
    x_rot = (
        np.cos((wind_direction - 270.0) * np.pi / 180.0) * x
        - np.sin((wind_direction - 270.0) * np.pi / 180.0) * y
    )
    y_rot = (
        np.sin((wind_direction - 270.0) * np.pi / 180.0) * x
        + np.cos((wind_direction - 270.0) * np.pi / 180.0) * y
    )

    if plot_lines:
        fig, ax = plt.subplots()
        for ii in range(n_turbs):
            ax.plot(
                x_rot[ii] * np.ones(2),
                [y_rot[ii] - D[ii] / 2, y_rot[ii] + D[ii] / 2],
                "k",
            )
        for ii in range(n_turbs):
            ax.text(x_rot[ii], y_rot[ii], "T%03d" % ii)
        ax.axis("equal")

    srt = np.argsort(x_rot)
    x_rot_srt = x_rot[srt]
    y_rot_srt = y_rot[srt]
    for ii in range(n_turbs):
        x0 = x_rot_srt[ii]
        y0 = y_rot_srt[ii]

        def wake_profile_ub_turbii(x):
            y = (y0 + D[ii]) + (x - x0) * wake_slope
            if isinstance(y, (float, np.float64, np.float32)):
                if x < (x0 + 0.01):
                    y = -np.Inf
            else:
                y[x < x0 + 0.01] = -np.Inf
            return y

        def wake_profile_lb_turbii(x):
            y = (y0 - D[ii]) - (x - x0) * wake_slope
            if isinstance(y, (float, np.float64, np.float32)):
                if x < (x0 + 0.01):
                    y = -np.Inf
            else:
                y[x < x0 + 0.01] = -np.Inf
            return y

        def determine_if_in_wake(xt, yt):
            return (yt < wake_profile_ub_turbii(xt)) & (yt > wake_profile_lb_turbii(xt))

        is_downstream[ii] = not any(
            [
                determine_if_in_wake(x_rot_srt[iii], y_rot_srt[iii])
                for iii in range(n_turbs)
            ]
        )

        if plot_lines:
            x1 = np.max(x_rot_srt) + 500.0
            ax.fill_between(
                [x0, x1, x1, x0],
                [
                    wake_profile_ub_turbii(x0 + 0.02),
                    wake_profile_ub_turbii(x1),
                    wake_profile_lb_turbii(x1),
                    wake_profile_lb_turbii(x0 + 0.02),
                ],
                alpha=0.1,
                color="k",
                edgecolor=None,
            )

    usrt = np.argsort(srt)
    is_downstream = [is_downstream[i] for i in usrt]
    turbs_downstream = list(np.where(is_downstream)[0])

    if plot_lines:
        ax.set_title("wind_direction = %03d" % wind_direction)
        ax.set_xlim([np.min(x_rot) - 500.0, x1])
        ax.set_ylim([np.min(y_rot) - 500.0, np.max(y_rot) + 500.0])
        ax.plot(
            x_rot[turbs_downstream], y_rot[turbs_downstream], "o", color="green",
        )

    return turbs_downstream



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import os

import numpy as np
from sklearn import neighbors

from ..utilities import Vec3


class FlowData:
    """
    FlowData objects represent a saved 3D flow from a FLORIS simulation
    or other data source.
    """

    # TODO handle none case, maybe default values apply like 0 origin and auto
    # determine spacing and dimensions
    def __init__(self, x, y, z, u, v, w, spacing=None, dimensions=None, origin=None):
        """
        Initialize FlowData object with coordinates, velocity fields,
        and meta-data.

        Args:
            x (np.array): Cartesian coordinate data.
            y (np.array): Cartesian coordinate data.
            z (np.array): Cartesian coordinate data.
            u (np.array): x-component of velocity.
            v (np.array): y-component of velocity.
            w (np.array): z-component of velocity.
            spacing (Vec3, optional): Spatial resolution.
                Defaults to None.
            dimensions (Vec3, optional): Named dimensions
                (e.g. x1, x2, x3). Defaults to None.
            origin (Vec3, optional): Coordinates of origin.
                Defaults to None.
        """

        self.x = x
        self.y = y
        self.z = z
        self.u = u
        self.v = v
        self.w = w

        self.spacing = spacing
        self.dimensions = dimensions
        self.origin = origin

        # Technically resolution is a restating of above, but it is useful to have
        self.resolution = Vec3(len(np.unique(x)), len(np.unique(y)), len(np.unique(z)))

    def save_as_vtk(self, filename):
        """
        Save FlowData Object to vtk format.

        Args:
            filename (str): Write-to path for vtk file.
        """
        n_points = self.dimensions.x1 * self.dimensions.x2 * self.dimensions.x3

        ln = "\n"
        vtk_file = open(filename, "w")
        vtk_file.write("# vtk DataFile Version 3.0" + ln)
        vtk_file.write("array.mean0D" + ln)
        vtk_file.write("ASCII" + ln)
        vtk_file.write("DATASET STRUCTURED_POINTS" + ln)
        vtk_file.write("DIMENSIONS {}".format(self.dimensions) + ln)
        vtk_file.write("ORIGIN {} {} {}".format(self.origin.x1, self.origin.x2, self.origin.x3) + ln)
        vtk_file.write("SPACING {}".format(self.spacing) + ln)
        vtk_file.write("POINT_DATA {}".format(n_points) + ln)
        vtk_file.write("FIELD attributes 1" + ln)
        vtk_file.write("UAvg 3 {} float".format(n_points) + ln)
        for u, v, w in zip(self.u, self.v, self.w):
            vtk_file.write("{}".format(Vec3(u, v, w)) + ln)

    @staticmethod
    def crop(ff, x_bnds, y_bnds, z_bnds):
        """
        Crop FlowData object to within stated bounds.

        Args:
            ff (:py:class:`~.tools.flow_data.FlowData`):
                FlowData object.
            x_bnds (iterable): Min and max of x-coordinate.
            y_bnds (iterable): Min and max of y-coordinate.
            z_bnds (iterable): Min and max of z-coordinate.

        Returns:
            (:py:class:`~.tools.flow_data.FlowData`):
            Cropped FlowData object.
        """

        map_values = (
            (ff.x > x_bnds[0])
            & (ff.x < x_bnds[1])
            & (ff.y > y_bnds[0])
            & (ff.y < y_bnds[1])
            & (ff.z > z_bnds[0])
            & (ff.z < z_bnds[1])
        )

        x = ff.x[map_values]
        y = ff.y[map_values]
        z = ff.z[map_values]

        #  Work out new dimensions
        dimensions = Vec3(len(np.unique(x)), len(np.unique(y)), len(np.unique(z)))

        # Work out origin
        origin = Vec3(
            ff.origin.x1 + np.min(x),
            ff.origin.x2 + np.min(y),
            ff.origin.x3 + np.min(z),
        )

        return FlowData(
            x - np.min(x),
            y - np.min(y),
            z - np.min(z),
            ff.u[map_values],
            ff.v[map_values],
            ff.w[map_values],
            spacing=ff.spacing,  # doesn't change
            dimensions=dimensions,
            origin=origin,
        )

        # Define a quick function for getting arbitrary points from sowfa

    def get_points_from_flow_data(self, x_points, y_points, z_points):
        """
        Return the u-value of a set of points from with a FlowData object.
        Use a simple nearest neighbor regressor to do internal interpolation.

        Args:
            x_points (np.array): Array of x-locations of points.
            y_points (np.array): Array of y-locations of points.
            z_points (np.array): Array of z-locations of points.

        Returns:
            np.array: Array of u-velocity at specified points.
        """
        # print(x_points,y_points,z_points)
        # X = np.column_stack([self.x, self.y, self.z])
        n_neighbors = 1
        knn = neighbors.KNeighborsRegressor(n_neighbors)
        # y_ = knn.fit(X, self.u)  # .predict(T)

        # Predict new points
        T = np.column_stack([x_points, y_points, z_points])
        return knn.predict(T)



from . import (
    yaw,
    layout,
    base_COE,
    optimization,
    layout_height,
    power_density,
    yaw_wind_rose,
    power_density_1D,
    yaw_wind_rose_parallel,
    derive_downstream_turbines,
)



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from typing import Any, Dict

from attrs import define, field
import numexpr as ne
import numpy as np

from floris.simulation import BaseModel
from floris.simulation import Farm
from floris.simulation import FlowField
from floris.simulation import Grid
from floris.simulation import Turbine
from floris.utilities import cosd, sind, tand


@define
class GaussVelocityDeficit(BaseModel):

    alpha: float = field(default=0.58)
    beta: float = field(default=0.077)
    ka: float = field(default=0.38)
    kb: float = field(default=0.004)
    model_string = "gauss"

    def prepare_function(
        self,
        grid: Grid,
        flow_field: FlowField,
        turbine: Turbine
    ) -> Dict[str, Any]:

        kwargs = dict(
            x=grid.x,
            y=grid.y,
            z=grid.z,
            reference_hub_height=turbine.hub_height,
            reference_rotor_diameter=turbine.rotor_diameter,
            u_initial=flow_field.u_initial,
            wind_veer=flow_field.wind_veer
        )
        return kwargs

    # @profile
    def function(
        self,
        x_i: np.ndarray,
        y_i: np.ndarray,
        z_i: np.ndarray,
        axial_induction_i: np.ndarray,
        deflection_field_i: np.ndarray,
        yaw_angle_i: np.ndarray,
        turbulence_intensity_i: np.ndarray,
        ct_i: np.ndarray,
        # enforces the use of the below as keyword arguments and adherence to the
        # unpacking of the results from prepare_function()
        *,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        reference_hub_height: float,
        reference_rotor_diameter: np.ndarray,
        u_initial: np.ndarray,
        wind_veer: float
    ) -> None:

        # yaw_angle is all turbine yaw angles for each wind speed
        # Extract and broadcast only the current turbine yaw setting
        # for all wind speeds

        # Opposite sign convention in this model
        yaw_angle = -1 * yaw_angle_i

        # Initialize the velocity deficit
        uR = u_initial * ct_i / ( 2.0 * (1 - np.sqrt(1 - ct_i) ) )
        u0 = u_initial * np.sqrt(1 - ct_i)

        # Initial lateral bounds
        sigma_z0 = reference_rotor_diameter * 0.5 * np.sqrt(uR / (u_initial + u0))
        sigma_y0 = sigma_z0 * cosd(yaw_angle) * cosd(wind_veer)


        # Compute the bounds of the near and far wake regions and a mask

        # Start of the near wake
        xR = x_i

        # Start of the far wake
        x0 = np.ones_like(u_initial)
        x0 *= reference_rotor_diameter * cosd(yaw_angle) * (1 + np.sqrt(1 - ct_i) )
        x0 /= np.sqrt(2) * (4 * self.alpha * turbulence_intensity_i + 2 * self.beta * (1 - np.sqrt(1 - ct_i) ) )
        x0 += x_i

        # Initialize the velocity deficit array
        velocity_deficit = np.zeros_like(u_initial)

        # Masks
        # When we have only an inequality, the current turbine may be applied its own wake in cases where numerical precision
        # cause in incorrect comparison. We've applied a small bump to avoid this. "0.1" is arbitrary but it is a small, non zero value.
        near_wake_mask = np.array(x > xR + 0.1) * np.array(x < x0)  # This mask defines the near wake; keeps the areas downstream of xR and upstream of x0
        far_wake_mask = np.array(x >= x0)

        # Compute the velocity deficit in the NEAR WAKE region
        # ONLY If there are points within the near wake boundary
        # TODO: for the turbinegrid, do we need to do this near wake calculation at all?
        #       same question for any grid with a resolution larger than the near wake region
        if np.sum(near_wake_mask):

            # Calculate the wake expansion
            near_wake_ramp_up = (x - xR) / (x0 - xR)  # This is a linear ramp from 0 to 1 from the start of the near wake to the start of the far wake.
            near_wake_ramp_down = (x0 - x) / (x0 - xR)  # Another linear ramp, but positive upstream of the far wake and negative in the far wake; 0 at the start of the far wake
            # near_wake_ramp_down = -1 * (near_wake_ramp_up - 1)  # TODO: this is equivalent, right?

            sigma_y = near_wake_ramp_down * 0.501 * reference_rotor_diameter * np.sqrt(ct_i / 2.0) + near_wake_ramp_up * sigma_y0
            sigma_y = sigma_y * np.array(x >= xR) + np.ones_like(sigma_y) * np.array(x < xR) * 0.5 * reference_rotor_diameter
            
            sigma_z = near_wake_ramp_down * 0.501 * reference_rotor_diameter * np.sqrt(ct_i / 2.0) + near_wake_ramp_up * sigma_z0
            sigma_z = sigma_z * np.array(x >= xR) + np.ones_like(sigma_z) * np.array(x < xR) * 0.5 * reference_rotor_diameter

            r, C = rC(
                wind_veer,
                sigma_y,
                sigma_z,
                y,
                y_i,
                deflection_field_i,
                z,
                reference_hub_height,
                ct_i,
                yaw_angle,
                reference_rotor_diameter
            )

            near_wake_deficit = gaussian_function(C, r, 1, np.sqrt(0.5))
            near_wake_deficit *= near_wake_mask

            velocity_deficit += near_wake_deficit


        # Compute the velocity deficit in the FAR WAKE region
        if np.sum(far_wake_mask):

            # Wake expansion in the lateral (y) and the vertical (z)
            ky = self.ka * turbulence_intensity_i + self.kb  # wake expansion parameters
            kz = self.ka * turbulence_intensity_i + self.kb  # wake expansion parameters
            sigma_y = (ky * (x - x0) + sigma_y0) * far_wake_mask + sigma_y0 * np.array(x < x0)
            sigma_z = (kz * (x - x0) + sigma_z0) * far_wake_mask + sigma_z0 * np.array(x < x0)

            r, C = rC(
                wind_veer,
                sigma_y,
                sigma_z,
                y,
                y_i,
                deflection_field_i,
                z,
                reference_hub_height,
                ct_i,
                yaw_angle,
                reference_rotor_diameter
            )

            far_wake_deficit = gaussian_function(C, r, 1, np.sqrt(0.5))
            far_wake_deficit *= far_wake_mask

            velocity_deficit += far_wake_deficit

        return velocity_deficit


# @profile
def rC(wind_veer, sigma_y, sigma_z, y, y_i, delta, z, HH, Ct, yaw, D):

    ## original
    # a = cosd(wind_veer) ** 2 / (2 * sigma_y ** 2) + sind(wind_veer) ** 2 / (2 * sigma_z ** 2)
    # b = -sind(2 * wind_veer) / (4 * sigma_y ** 2) + sind(2 * wind_veer) / (4 * sigma_z ** 2)
    # c = sind(wind_veer) ** 2 / (2 * sigma_y ** 2) + cosd(wind_veer) ** 2 / (2 * sigma_z ** 2)
    # r = (
    #     a * (y - y_i - delta) ** 2
    #     - 2 * b * (y - y_i - delta) * (z - HH)
    #     + c * (z - HH) ** 2
    # )
    # C = 1 - np.sqrt( np.clip(1 - (Ct * cosd(yaw) / (8.0 * sigma_y * sigma_z / D ** 2)), 0.0, 1.0) )

    ## Precalculate some parts
    # twox_sigmay_2 = 2 * sigma_y ** 2
    # twox_sigmaz_2 = 2 * sigma_z ** 2
    # a = cosd(wind_veer) ** 2 / (twox_sigmay_2) + sind(wind_veer) ** 2 / (twox_sigmaz_2)
    # b = -sind(2 * wind_veer) / (2 * twox_sigmay_2) + sind(2 * wind_veer) / (2 * twox_sigmaz_2)
    # c = sind(wind_veer) ** 2 / (twox_sigmay_2) + cosd(wind_veer) ** 2 / (twox_sigmaz_2)
    # delta_y = y - y_i - delta
    # delta_z = z - HH
    # r = (a * (delta_y ** 2) - 2 * b * (delta_y) * (delta_z) + c * (delta_z ** 2))
    # C = 1 - np.sqrt( np.clip(1 - (Ct * cosd(yaw) / ( 8.0 * sigma_y * sigma_z / (D * D) )), 0.0, 1.0) )

    ## Numexpr
    wind_veer = np.deg2rad(wind_veer)
    a = ne.evaluate("cos(wind_veer) ** 2 / (2 * sigma_y ** 2) + sin(wind_veer) ** 2 / (2 * sigma_z ** 2)")
    b = ne.evaluate("-sin(2 * wind_veer) / (4 * sigma_y ** 2) + sin(2 * wind_veer) / (4 * sigma_z ** 2)")
    c = ne.evaluate("sin(wind_veer) ** 2 / (2 * sigma_y ** 2) + cos(wind_veer) ** 2 / (2 * sigma_z ** 2)")
    r = ne.evaluate("a * ( (y - y_i - delta) ** 2) - 2 * b * (y - y_i - delta) * (z - HH) + c * ((z - HH) ** 2)")
    d = np.clip(1 - (Ct * cosd(yaw) / ( 8.0 * sigma_y * sigma_z / (D * D) )), 0.0, 1.0)
    C = ne.evaluate("1 - sqrt(d)")
    return r, C

def mask_upstream_wake(mesh_y_rotated, x_coord_rotated, y_coord_rotated, turbine_yaw):
    yR = mesh_y_rotated - y_coord_rotated
    xR = yR * tand(turbine_yaw) + x_coord_rotated
    return xR, yR

def gaussian_function(C, r, n, sigma):
    return C * np.exp(-1 * r ** n / (2 * sigma ** 2))



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


"""
The :py:obj:`floris.tools` package contains the modules used to drive
FLORIS simulations and perform studies in various areas of research and
analysis.

All modules can be imported with

    >>> import floris.tools

The ``__init__.py`` file enables the import of all modules in this
package so any additional modules should be included there.

Examples:
    >>> import floris.tools

    >>> dir(floris.tools)
    ['__builtins__', '__cached__', '__doc__', '__file__', '__loader__',
    '__name__', '__package__', '__path__', '__spec__', 'cut_plane',
    'energy_ratio', 'floris_interface', 'flow_data',
    'layout_functions', 'optimization', 'plotting', 'power_rose',
    'rews', 'sowfa_utilities', 'visualization', 'wind_rose']
"""

from .floris_interface import FlorisInterface
from .visualization import visualize_cut_plane, visualize_quiver, plot_turbines_with_fi, plot_rotor_values
from .wind_rose import WindRose

# from floris.tools import (
    # cut_plane,
    # energy_ratio,
    # energy_ratio_single,
    # floris_interface,
    # flow_data,
    # interface_utilities,
    # layout_functions,
    # optimization,
    # plotting,
    # power_rose,
    # rews,
    # sowfa_utilities,
    # visualization,
    # wind_rose,
# )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import numpy as np

from .base_velocity_deflection import VelocityDeflection


class Curl(VelocityDeflection):
    """
    Stand-in class for the curled wake model. Wake deflection with the curl
    model is handled inherently in the wake velocity portion of the model.
    Passes zeros for deflection values. See
    :cite:`cdm-martinez2019aerodynamics` for additional info on the curled wake
    model.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: cdm-
    """

    def __init__(self, parameter_dictionary):
        """
        See super-class for initialization details. See
        :py:class:`floris.simulation.wake_velocity.curl` for details on
        `parameter_dictionary`.

        Args:
            parameter_dictionary (dict): Model-specific parameters.
        """
        super().__init__(parameter_dictionary)
        self.model_string = "curl"

    def function(
        self, x_locations, y_locations, z_locations, turbine, coord, flow_field
    ):
        """
        Passes zeros for wake deflection as deflection is inherently handled in
        the wake velocity portion of the curled wake model.

        Args:
            x_locations (np.array): An array of floats that contains the
                streamwise direction grid coordinates of the flow field
                domain (m).
            y_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the direction normal to
                x and parallel to the ground (m).
            z_locations (np.array): An array of floats that contains the grid
                coordinates of the flow field domain in the vertical
                direction (m).
            turbine (:py:obj:`floris.simulation.turbine`): Object that
                represents the turbine creating the wake.
            coord (:py:obj:`floris.utilities.Vec3`): Object containing
                the coordinate of the turbine creating the wake (m).
            flow_field (:py:class:`floris.simulation.flow_field`): Object
                containing the flow field information for the wind farm.

        Returns:
            np.array: Zeros the same size as the flow field grid points.
        """
        return np.zeros(np.shape(x_locations))



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize

from .optimization import Optimization


class PowerDensityOptimization1D(Optimization):
    """
    PowerDensityOptimization1D is a subclass of the
    :py:class:`~.tools.optimization.scipy.optimization.Optimization` class
    that performs layout optimization in 1 dimension. TODO: What is this single
    dimension?
    """

    def __init__(
        self,
        fi,
        wd,
        ws,
        freq,
        AEP_initial,
        x0=None,
        bnds=None,
        min_dist=None,
        opt_method="SLSQP",
        opt_options=None,
    ):
        """
        Instantiate PowerDensityOptimization1D object with a FlorisInterface
        object and assigns parameter values.

        Args:
            fi (:py:class:`floris.tools.floris_interface.FlorisInterface`):
                Interface used to interact with the Floris object.
            wd (np.array): An array of wind directions (deg).
            ws (np.array): An array of wind speeds (m/s).
            freq (np.array): An array of the frequencies of occurance
                correponding to each pair of wind direction and wind speed
                values.
            AEP_initial (float): The initial Annual Energy
                Production used for normalization in the optimization (Wh)
                (TODO: Is Watt-hours the correct unit?).
            x0 (iterable, optional): The initial turbine locations,
                ordered by x-coordinate and then y-coordiante
                (ie. [x1, x2, ..., xn, y1, y2, ..., yn]) (m). If none are
                provided, x0 initializes to the current turbine locations.
                Defaults to None.
            bnds (iterable, optional): Bounds for the optimization
                variables (pairs of min/max values for each variable (m)). If
                none are specified, they are set to some example values (TODO:
                what is the significance of these example values?). Defaults to
                None.
            min_dist (float, optional): The minimum distance to be
                maintained between turbines during the optimization (m). If not
                specified, initializes to 2 rotor diameters. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to 'SLSQP'.
            opt_options (dict, optional): Optimization options used by
                scipy.optimize.minize. If none are specified, they are set to
                {'maxiter': 100, 'disp': True, 'iprint': 2, 'ftol': 1e-9}.
                Defaults to None.
        """
        super().__init__(fi)
        self.epsilon = np.finfo(float).eps
        self.counter = 0

        if opt_options is None:
            self.opt_options = {"maxiter": 100, "disp": True, "iprint": 2, "ftol": 1e-9}

        self.reinitialize_opt(
            wd=wd,
            ws=ws,
            freq=freq,
            AEP_initial=AEP_initial,
            x0=x0,
            bnds=bnds,
            min_dist=min_dist,
            opt_method=opt_method,
            opt_options=opt_options,
        )

    def _PowDens_opt(self, optVars):
        locs = optVars[0 : self.nturbs]
        locs_unnorm = [
            self._unnorm(valx, self.bndx_min, self.bndx_max) for valx in locs
        ]
        turb_controls = [
            optVars[self.nturbs + i * self.nturbs : 2 * self.nturbs + i * self.nturbs]
            for i in range(len(self.wd))
        ]
        turb_controls_unnorm = [
            self._unnorm(yaw, self.yaw_min, self.yaw_max) for yaw in turb_controls
        ]

        self._change_coordinates(locs_unnorm)

        for i, turbine in enumerate(self.fi.floris.farm.turbine_map.turbines):
            turbine.yaw_angle = turb_controls_unnorm[0][i]

        layout_dist = self._avg_dist(locs)
        # AEP_sum = self._AEP_single_wd(self.wd[0], self.ws[0])
        # print('AEP ratio: ', AEP_sum/self.AEP_initial)

        return layout_dist / self.layout_dist_initial

    def _avg_dist(self, locs):
        dist = []
        for i in range(len(locs) - 1):
            dist.append(locs[i + 1] - locs[i])

        return np.mean(dist)

    def _change_coordinates(self, locs):
        # Parse the layout coordinates
        layout_x = locs
        layout_y = [coord.x2 for coord in self.fi.floris.farm.turbine_map.coords]
        layout_array = [layout_x, layout_y]

        # Update the turbine map in floris
        self.fi.reinitialize_flow_field(layout_array=layout_array)

    def _set_opt_bounds(self):
        # self.bnds = [(0.0, 1.0) for _ in range(2*self.nturbs)]
        self.bnds = [
            (0.0, 0.0),
            (0.083333, 0.25),
            (0.166667, 0.5),
            (0.25, 0.75),
            (0.33333, 1.0),
            (0.0, 1.0),
            (0.0, 1.0),
            (0.0, 1.0),
            (0.0, 1.0),
            (0.0, 1.0),
        ]

    def _AEP_single_wd(self, wd, ws):
        self.fi.reinitialize_flow_field(wind_direction=wd, wind_speed=ws)
        self.fi.calculate_wake()

        turb_powers = [turbine.power for turbine in self.fi.floris.farm.turbines]
        return np.sum(turb_powers) * self.freq[0] * 8760

    def _AEP_constraint(self, optVars):
        locs = optVars[0 : self.nturbs]
        locs_unnorm = [
            self._unnorm(valx, self.bndx_min, self.bndx_max) for valx in locs
        ]
        turb_controls = [
            optVars[self.nturbs + i * self.nturbs : 2 * self.nturbs + i * self.nturbs]
            for i in range(len(self.wd))
        ]
        turb_controls_unnorm = [
            self._unnorm(yaw, self.yaw_min, self.yaw_max) for yaw in turb_controls
        ]

        for i, turbine in enumerate(self.fi.floris.farm.turbine_map.turbines):
            turbine.yaw_angle = turb_controls_unnorm[0][i]

        self._change_coordinates(locs_unnorm)

        return (
            self._AEP_single_wd(self.wd[0], self.ws[0]) / self.AEP_initial - 1
        ) * 1000000.0

    def _space_constraint(self, x_in, min_dist):
        x = np.nan_to_num(x_in[0 : self.nturbs])
        y = np.nan_to_num(x_in[self.nturbs :])

        dist = [
            np.sqrt((x[i] - x[j]) ** 2 + (y[i] - y[j]) ** 2)
            for i in range(self.nturbs)
            for j in range(self.nturbs)
            if i != j
        ]

        return np.min(dist) - self._norm(min_dist, self.bndx_min, self.bndx_max)

    def _generate_constraints(self):
        tmp1 = {
            "type": "ineq",
            "fun": lambda x, *args: self._space_constraint(x, self.min_dist),
            "args": (self.min_dist,),
        }
        tmp2 = {"type": "ineq", "fun": lambda x, *args: self._AEP_constraint(x)}

        self.cons = [tmp1, tmp2]

    def _optimize(self):
        self.residual_plant = minimize(
            self._PowDens_opt,
            self.x0,
            method=self.opt_method,
            bounds=self.bnds,
            constraints=self.cons,
            options=self.opt_options,
        )

        opt_results = self.residual_plant.x

        return opt_results

    def optimize(self):
        """
        This method finds the optimized layout of wind turbines for power
        production given the provided frequencies of occurance of wind
        conditions (wind speed, direction).

        Returns:
            opt_locs (iterable): A list of the optimized x, y locations of each
            turbine (m).
        """
        print("=====================================================")
        print("Optimizing turbine layout...")
        print("Number of parameters to optimize = ", len(self.x0))
        print("=====================================================")

        opt_vars_norm = self._optimize()

        print("Optimization complete.")

        opt_locs = [
            self._unnorm(valx, self.bndx_min, self.bndx_max)
            for valx in opt_vars_norm[0 : self.nturbs]
        ]

        opt_yaw = [
            self._unnorm(yaw, self.yaw_min, self.yaw_max)
            for yaw in opt_vars_norm[self.nturbs :]
        ]

        return [opt_locs, opt_yaw]

    def reinitialize_opt(
        self,
        wd=None,
        ws=None,
        freq=None,
        AEP_initial=None,
        x0=None,
        bnds=None,
        min_dist=None,
        yaw_lims=None,
        opt_method=None,
        opt_options=None,
    ):
        """
        This method reinitializes any optimization parameters that are
        specified. Otherwise, the current parameter values are kept.

        Args:
            wd (np.array): An array of wind directions (deg). Defaults to None.
            ws (np.array): An array of wind speeds (m/s). Defaults to None.
            freq (np.array): An array of the frequencies of occurance
                correponding to each pair of wind direction and wind speed
                values. Defaults to None.
            AEP_initial (float): The initial Annual Energy
                Production used for normalization in the optimization (Wh)
                (TODO: Is Watt-hours the correct unit?). Defaults to None.
            x0 (iterable, optional): The initial turbine locations,
                ordered by x-coordinate and then y-coordiante
                (ie. [x1, x2, ..., xn, y1, y2, ..., yn]) (m). If none are
                provided, x0 initializes to the current turbine locations.
                Defaults to None.
            bnds (iterable, optional): Bounds for the optimization
                variables (pairs of min/max values for each variable (m)). If
                none are specified, they are set to some example values (TODO:
                what is the significance of these example values?). Defaults to
                None.
            min_dist (float, optional): The minimum distance to be
                maintained between turbines during the optimization (m). If not
                specified, initializes to 2 rotor diameters. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to None.
            opt_options (dict, optional): Optimization options used by
                scipy.optimize.minize. Defaults to None.
        """
        # if boundaries is not None:
        #     self.boundaries = boundaries
        #     self.bndx_min = np.min([val[0] for val in boundaries])
        #     self.bndy_min = np.min([val[1] for val in boundaries])
        #     self.boundaries_norm = [[self._norm(val[0], self.bndx_min, \
        #                           self.bndx_max)] for val in self.boundaries]
        self.bndx_min = np.min(
            [coord.x1 for coord in self.fi.floris.farm.turbine_map.coords]
        )
        self.bndx_max = np.max(
            [coord.x1 for coord in self.fi.floris.farm.turbine_map.coords]
        )
        if yaw_lims is not None:
            self.yaw_min = yaw_lims[0]
            self.yaw_max = yaw_lims[1]
        else:
            self.yaw_min = 0.0
            self.yaw_max = 20.0
        if wd is not None:
            self.wd = wd
        if ws is not None:
            self.ws = ws
        if freq is not None:
            self.freq = freq
        if AEP_initial is not None:
            self.AEP_initial = AEP_initial
        else:
            self.AEP_initial = self.fi.get_farm_AEP(self.wd, self.ws, self.freq)
        if x0 is not None:
            self.x0 = x0
        else:
            self.x0 = [
                self._norm(coord.x1, self.bndx_min, self.bndx_max)
                for coord in self.fi.floris.farm.turbine_map.coords
            ] + [0.0] * self.nturbs

        if bnds is not None:
            self.bnds = bnds
        else:
            self._set_opt_bounds()
        if min_dist is not None:
            self.min_dist = min_dist
        else:
            self.min_dist = 2 * self.fi.floris.farm.turbines[0].rotor_diameter
        if opt_method is not None:
            self.opt_method = opt_method
        if opt_options is not None:
            self.opt_options = opt_options

        self._generate_constraints()
        # self.layout_dist_initial = np.max(self.x0[0:self.nturbs]) \
        #    - np.min(self.x0[0:self.nturbs])
        self.layout_dist_initial = self._avg_dist(self.x0[0 : self.nturbs])
        # print('initial dist: ', self.layout_dist_initial)

    def plot_layout_opt_results(self):
        """
        This method plots the original and new locations of the turbines in a
        wind farm after layout optimization.
        """
        locsx_old = [
            self._unnorm(valx, self.bndx_min, self.bndx_max)
            for valx in self.x0[0 : self.nturbs]
        ]
        locsy_old = self.fi.layout_y
        locsx = [
            self._unnorm(valx, self.bndx_min, self.bndx_max)
            for valx in self.residual_plant.x[0 : self.nturbs]
        ]
        locsy = self.fi.layout_y

        plt.figure(figsize=(9, 6))
        fontsize = 16
        plt.plot(locsx_old, locsy_old, "ob")
        plt.plot(locsx, locsy, "or")
        # plt.title('Layout Optimization Results', fontsize=fontsize)
        plt.xlabel("x (m)", fontsize=fontsize)
        plt.ylabel("y (m)", fontsize=fontsize)
        plt.axis("equal")
        plt.grid()
        plt.tick_params(which="both", labelsize=fontsize)
        plt.legend(
            ["Old locations", "New locations"],
            loc="lower center",
            bbox_to_anchor=(0.5, 1.01),
            ncol=2,
            fontsize=fontsize,
        )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import numpy as np

from shapely.geometry import Point, Polygon


# def discontinuous_grid(nrows,ncols,farm_width,farm_height,shear,rotation,center_x,center_y,boundary_mult,boundary_x,boundary_y):
def discontinuous_grid(
    nrows,
    ncols,
    farm_width,
    farm_height,
    shear,
    rotation,
    center_x,
    center_y,
    shrink_boundary,
    boundary_x,
    boundary_y,
    eps=1e-3,
):
    """
    Map from grid design variables to turbine x and y locations. Includes integer design variables and the formulation
    results in a discontinous design space.

    TODO: shrink_boundary doesn't work well with concave boundaries, or with boundary angles less than 90 deg

    Args:
        nrows (Int): number of rows in the grid.
        ncols (Int): number of columns in the grid.
        farm_width (Float): total grid width (before shear).
        farm_height (Float): total grid height.
        shear (Float): grid shear (rad).
        rotation (Float): rotation about grid center (rad).
        center_x (Float): location of grid x center.
        center_y (Float): location of grid y center.
        shrink_boundary (Float): how much to shrink the boundary that the grid can occupy.
        boundary_x (Array(Float)): x boundary points.
        boundary_y (Array(Float)): y boundary points.

    Returns:
        grid_x (Array(Float)): turbine x locations.
        grid_y (Array(Float)): turbine y locations.
    """
    # create grid
    nrows = int(nrows)
    ncols = int(ncols)
    xlocs = np.linspace(0.0, farm_width, ncols)
    ylocs = np.linspace(0.0, farm_height, nrows)
    y_spacing = ylocs[1] - ylocs[0]
    nturbs = nrows * ncols
    grid_x = np.zeros(nturbs)
    grid_y = np.zeros(nturbs)
    turb = 0
    for i in range(nrows):
        for j in range(ncols):
            grid_x[turb] = xlocs[j] + float(i) * y_spacing * np.tan(shear)
            grid_y[turb] = ylocs[i]
            turb += 1

    # rotate
    grid_x, grid_y = (
        np.cos(rotation) * grid_x - np.sin(rotation) * grid_y,
        np.sin(rotation) * grid_x + np.cos(rotation) * grid_y,
    )

    # move center of grid
    grid_x = (grid_x - np.mean(grid_x)) + center_x
    grid_y = (grid_y - np.mean(grid_y)) + center_y

    # arrange the boundary

    # boundary = np.zeros((len(boundary_x),2))
    # boundary[:,0] = boundary_x[:]
    # boundary[:,1] = boundary_y[:]
    # poly = Polygon(boundary)
    # centroid = poly.centroid

    # boundary[:,0] = (boundary_x[:]-centroid.x)*boundary_mult + centroid.x
    # boundary[:,1] = (boundary_y[:]-centroid.y)*boundary_mult + centroid.y
    # poly = Polygon(boundary)

    boundary = np.zeros((len(boundary_x), 2))
    boundary[:, 0] = boundary_x[:]
    boundary[:, 1] = boundary_y[:]
    poly = Polygon(boundary)

    if shrink_boundary != 0.0:
        nBounds = len(boundary_x)
        for i in range(nBounds):
            point = Point(boundary_x[i] + eps, boundary_y[i])
            if poly.contains(point) is True or poly.touches(point) is True:
                boundary[i, 0] = boundary_x[i] + shrink_boundary
            else:
                boundary[i, 0] = boundary_x[i] - shrink_boundary

            point = Point(boundary_x[i], boundary_y[i] + eps)
            if poly.contains(point) is True or poly.touches(point) is True:
                boundary[i, 1] = boundary_y[i] + shrink_boundary
            else:
                boundary[i, 1] = boundary_y[i] - shrink_boundary

        poly = Polygon(boundary)

    # get rid of points outside of boundary
    index = 0
    for i in range(len(grid_x)):
        point = Point(grid_x[index], grid_y[index])
        if poly.contains(point) is False and poly.touches(point) is False:
            grid_x = np.delete(grid_x, index)
            grid_y = np.delete(grid_y, index)
        else:
            index += 1

    return grid_x, grid_y


def place_boundary_turbines(n_boundary_turbs, start, boundary_x, boundary_y):
    """
    Place turbines equally spaced traversing the perimiter if the wind farm along the boundary

    Args:
    n_boundary_turbs (Int): number of turbines to be placed on the boundary
    start (Float): where the first turbine should be placed
    boundary_x (Array(Float)): x boundary points
    boundary_y (Array(Float)): y boundary points

    Returns
    layout_x (Array(Float)): turbine x locations
    layout_y (Array(Float)): turbine y locations
    """

    # check if the boundary is closed, correct if not
    if boundary_x[-1] != boundary_x[0] or boundary_y[-1] != boundary_y[0]:
        boundary_x = np.append(boundary_x, boundary_x[0])
        boundary_y = np.append(boundary_y, boundary_y[0])

    # make the boundary
    boundary = np.zeros((len(boundary_x), 2))
    boundary[:, 0] = boundary_x[:]
    boundary[:, 1] = boundary_y[:]
    poly = Polygon(boundary)
    perimeter = poly.length

    # get the flattened turbine locations
    spacing = perimeter / float(n_boundary_turbs)
    flattened_locs = np.linspace(start, perimeter + start - spacing, n_boundary_turbs)

    # set all of the flattened values between 0 and the perimeter
    for i in range(n_boundary_turbs):
        while flattened_locs[i] < 0.0:
            flattened_locs[i] += perimeter
        if flattened_locs[i] > perimeter:
            flattened_locs[i] = flattened_locs[i] % perimeter

    # place the turbines around the perimeter
    nBounds = len(boundary_x)
    layout_x = np.zeros(n_boundary_turbs)
    layout_y = np.zeros(n_boundary_turbs)

    lenBound = np.zeros(nBounds - 1)
    for i in range(nBounds - 1):
        lenBound[i] = Point(boundary[i]).distance(Point(boundary[i + 1]))
    for i in range(n_boundary_turbs):
        for j in range(nBounds - 1):
            if flattened_locs[i] < sum(lenBound[0 : j + 1]):
                layout_x[i] = (
                    boundary_x[j]
                    + (boundary_x[j + 1] - boundary_x[j])
                    * (flattened_locs[i] - sum(lenBound[0:j]))
                    / lenBound[j]
                )
                layout_y[i] = (
                    boundary_y[j]
                    + (boundary_y[j + 1] - boundary_y[j])
                    * (flattened_locs[i] - sum(lenBound[0:j]))
                    / lenBound[j]
                )
                break

    return layout_x, layout_y


def boundary_grid(
    n_boundary_turbs,
    start,
    nrows,
    ncols,
    farm_width,
    farm_height,
    shear,
    rotation,
    center_x,
    center_y,
    shrink_boundary,
    boundary_x,
    boundary_y,
    eps=1e-3,
):
    """
    Place turbines equally spaced traversing the perimiter if the wind farm along the boundary

    Args:
    n_boundary_turbs,start: boundary variables
    nrows,ncols,farm_width,farm_height,shear,rotation,center_x,center_y,shrink_boundary,eps: grid variables
    boundary_x,boundary_y: boundary points

    Returns
    layout_x (Array(Float)): turbine x locations
    layout_y (Array(Float)): turbine y locations
    """

    boundary_turbines_x, boundary_turbines_y = place_boundary_turbines(
        n_boundary_turbs, start, boundary_x, boundary_y
    )
    grid_turbines_x, grid_turbines_y = discontinuous_grid(
        nrows,
        ncols,
        farm_width,
        farm_height,
        shear,
        rotation,
        center_x,
        center_y,
        shrink_boundary,
        boundary_x,
        boundary_y,
        eps=eps,
    )

    layout_x = np.append(boundary_turbines_x, grid_turbines_x)
    layout_y = np.append(boundary_turbines_y, grid_turbines_y)

    return layout_x, layout_y


class BoundaryGrid:
    """
    Parameterize the wind farm layout with a grid or the boundary grid method
    """

    def __init__(self, fi):
        """
        Initializes a BoundaryGrid object by assigning a
        FlorisInterface object.

        Args:
            fi (:py:class:`~.tools.floris_interface.FlorisInterface`):
                Interface used to interact with the Floris object.
        """
        self.fi = fi

        self.n_boundary_turbs = 0
        self.start = 0.0
        self.nrows = 0
        self.ncols = 0
        self.farm_width = 0.0
        self.farm_height = 0.0
        self.shear = 0.0
        self.rotation = 0.0
        self.center_x = 0.0
        self.center_y = 0.0
        self.shrink_boundary = 0.0
        self.boundary_x = np.array([])
        self.boundary_y = np.array([])
        self.eps = 1e-3

    def reinitialize_bg(
        self,
        n_boundary_turbs=None,
        start=None,
        nrows=None,
        ncols=None,
        farm_width=None,
        farm_height=None,
        shear=None,
        rotation=None,
        center_x=None,
        center_y=None,
        shrink_boundary=None,
        boundary_x=None,
        boundary_y=None,
        eps=None,
    ):

        if n_boundary_turbs is not None:
            self.n_boundary_turbs = n_boundary_turbs
        if start is not None:
            self.start = start
        if nrows is not None:
            self.nrows = nrows
        if ncols is not None:
            self.ncols = ncols
        if farm_width is not None:
            self.farm_width = farm_width
        if farm_height is not None:
            self.farm_height = farm_height
        if shear is not None:
            self.shear = shear
        if rotation is not None:
            self.rotation = rotation
        if center_x is not None:
            self.center_x = center_x
        if center_y is not None:
            self.center_y = center_y
        if shrink_boundary is not None:
            self.shrink_boundary = shrink_boundary
        if boundary_x is not None:
            self.boundary_x = boundary_x
        if boundary_y is not None:
            self.boundary_y = boundary_y
        if eps is not None:
            self.eps = eps

    def reinitialize_xy(self):

        layout_x, layout_y = boundary_grid(
            self.n_boundary_turbs,
            self.start,
            self.nrows,
            self.ncols,
            self.farm_width,
            self.farm_height,
            self.shear,
            self.rotation,
            self.center_x,
            self.center_y,
            self.shrink_boundary,
            self.boundary_x,
            self.boundary_y,
            eps=self.eps,
        )

        self.fi.reinitialize_flow_field(layout_array=(layout_x, layout_y))


if __name__ == "__main__":

    nrows = 10
    ncols = 10
    farm_width = 600
    farm_height = 600
    shear = np.deg2rad(10)
    rotation = np.deg2rad(30)
    center_x = 250
    center_y = 300
    boundary_mult = 0.6
    shrink_boundary = 10.0
    # boundary_x = np.array([-300.0,-100.0,100.0,100.0,0.0]) + 2200.0
    # boundary_y = np.array([-100.0,100.0,140.0,-100.0,0.0]) + 300.0
    boundary_x = np.array(
        [0.0, 100.0, 100.0, 200.0, 200.0, 300.0, 300.0, 400.0, 400.0, 500.0, 500.0, 0.0]
    )
    boundary_y = np.array(
        [
            500.0,
            500.0,
            400.0,
            400.0,
            300.0,
            300.0,
            200.0,
            200.0,
            100.0,
            100.0,
            600.0,
            600.0,
        ]
    )
    # x,y = discontinuous_grid(nrows,ncols,farm_width,farm_height,shear,rotation,center_x,center_y,boundary_mult,boundary_x,boundary_y)
    x, y = discontinuous_grid(
        nrows,
        ncols,
        farm_width,
        farm_height,
        shear,
        rotation,
        center_x,
        center_y,
        shrink_boundary,
        boundary_x,
        boundary_y,
    )

    n_boundary_turbs = 25
    start = 1000.0
    layout_x, layout_y = place_boundary_turbines(
        n_boundary_turbs, start, boundary_x, boundary_y
    )

    bx = np.append(boundary_x, boundary_x[0])
    by = np.append(boundary_y, boundary_y[0])

    boundary = np.zeros((len(bx), 2))
    boundary[:, 0] = bx[:]
    boundary[:, 1] = by[:]
    poly = Polygon(boundary)

    # centroid = poly.centroid
    # new_bx = (bx[:]-centroid.x)*boundary_mult + centroid.x
    # new_by = (by[:]-centroid.y)*boundary_mult + centroid.y

    nBounds = len(bx)
    new_bx = np.zeros(len(bx))
    new_by = np.zeros(len(by))
    eps = 1e-3
    for i in range(nBounds):
        point = Point(bx[i] + eps, by[i])
        if poly.contains(point) is True or poly.touches(point) is True:
            new_bx[i] = bx[i] + shrink_boundary
        else:
            new_bx[i] = bx[i] - shrink_boundary

        point = Point(bx[i], by[i] + eps)
        if poly.contains(point) is True or poly.touches(point) is True:
            new_by[i] = by[i] + shrink_boundary
        else:
            new_by[i] = by[i] - shrink_boundary

    import matplotlib.pyplot as plt

    nx, ny = boundary_grid(
        n_boundary_turbs,
        start,
        nrows,
        ncols,
        farm_width,
        farm_height,
        shear,
        rotation,
        center_x,
        center_y,
        shrink_boundary,
        boundary_x,
        boundary_y,
    )

    plt.plot(bx, by)
    plt.plot(new_bx, new_by)
    plt.plot(x, y, "o")
    plt.plot(layout_x, layout_y, "o")
    plt.axis("equal")

    plt.figure(2)
    plt.plot(bx, by)
    plt.plot(nx, ny, "o")
    plt.axis("equal")

    plt.show()



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

from __future__ import annotations

import json
from pathlib import Path
import yaml

import floris.logging_manager as logging_manager
from floris.type_dec import FromDictMixin
from floris.simulation import (
    Farm,
    WakeModelManager,
    FlowField,
    Turbine,
    Grid,
    TurbineGrid,
    FlowFieldGrid,
    FlowFieldPlanarGrid,
    sequential_solver,
    cc_solver,
    full_flow_sequential_solver,
    full_flow_cc_solver
)
from attrs import define, field


@define
class Floris(logging_manager.LoggerBase, FromDictMixin):
    """
    Top-level class that describes a Floris model and initializes the
    simulation. Use the :py:class:`~.simulation.farm.Farm` attribute to
    access other objects within the model.
    """

    logging: dict = field(converter=dict)
    solver: dict = field(converter=dict)
    wake: WakeModelManager = field(converter=WakeModelManager.from_dict)
    turbine: Turbine = field(converter=Turbine.from_dict)
    farm: Farm = field(converter=Farm.from_dict)
    flow_field: FlowField = field(converter=FlowField.from_dict)

    # These fields are included to appease the requirement that all inputs must
    # be mapped to a field in the class. They are not used in FLORIS.
    name: str  = field(converter=str)
    description: str = field(converter=str)
    floris_version: str = field(converter=str)

    grid: Grid = field(init=False)

    def __attrs_post_init__(self) -> None:

        # Initialize farm quanitities that depend on other objects
        self.farm.construct_coordinates(self.flow_field.reference_wind_height)
        self.farm.set_yaw_angles(self.flow_field.n_wind_directions, self.flow_field.n_wind_speeds)

        if self.solver["type"] == "turbine_grid":
            self.grid = TurbineGrid(
                turbine_coordinates=self.farm.coordinates,
                reference_turbine_diameter=self.turbine.rotor_diameter,
                wind_directions=self.flow_field.wind_directions,
                wind_speeds=self.flow_field.wind_speeds,
                grid_resolution=self.solver["turbine_grid_points"],
            )
        elif self.solver["type"] == "flow_field_grid":
            self.grid = FlowFieldGrid(
                turbine_coordinates=self.farm.coordinates,
                reference_turbine_diameter=self.turbine.rotor_diameter,
                wind_directions=self.flow_field.wind_directions,
                wind_speeds=self.flow_field.wind_speeds,
                grid_resolution=self.solver["flow_field_grid_points"],
            )
        elif self.solver["type"] == "flow_field_planar_grid":
            self.grid = FlowFieldPlanarGrid(
                turbine_coordinates=self.farm.coordinates,
                reference_turbine_diameter=self.turbine.rotor_diameter,
                wind_directions=self.flow_field.wind_directions,
                wind_speeds=self.flow_field.wind_speeds,
                normal_vector=self.solver["normal_vector"],
                planar_coordinate=self.solver["planar_coordinate"],
                grid_resolution=self.solver["flow_field_grid_points"],
                x1_bounds=self.solver["flow_field_bounds"][0],
                x2_bounds=self.solver["flow_field_bounds"][1],
            )
        else:
            raise ValueError(
                f"Supported solver types are [turbine_grid, flow_field_grid], but type given was {self.solver['type']}"
            )

        # Configure logging
        logging_manager.configure_console_log(
            self.logging["console"]["enable"],
            self.logging["console"]["level"],
        )
        logging_manager.configure_file_log(
            self.logging["file"]["enable"],
            self.logging["file"]["level"],
        )
    

    # @profile
    def steady_state_atmospheric_condition(self):

        # Initialize field quanitities; doing this immediately prior to doing
        # the calculation step allows for manipulating inputs in a script
        # without changing the data structures
        self.flow_field.initialize_velocity_field(self.grid)

        # Initialize farm quantities
        self.farm.initialize(self.grid.sorted_indices)

        vel_model = self.wake.model_strings["velocity_model"]

        # <<interface>>
        # start = time.time()

        if vel_model=="cc":
            elapsed_time = cc_solver(
                self.farm,
                self.flow_field,
                self.turbine,
                self.grid,
                self.wake
            )
        else:
            elapsed_time = sequential_solver(
                self.farm,
                self.flow_field,
                self.turbine,
                self.grid,
                self.wake
            )
        # end = time.time()
        # elapsed_time = end - start

        self.grid.finalize()
        self.flow_field.finalize(self.grid.unsorted_indices)
        self.farm.finalize(self.grid.unsorted_indices)
        return elapsed_time

    def solve_for_viz(self):
        # Do the calculation with the TurbineGrid for a single wind speed
        # and wind direction and 1 point on the grid. Then, use the result
        # to construct the full flow field grid.
        # This function call should be for a single wind direction and wind speed
        # since the memory consumption is very large.

        self.flow_field.initialize_velocity_field(self.grid)

        vel_model = self.wake.model_strings["velocity_model"]

        if vel_model=="cc":
            full_flow_cc_solver(self.farm, self.flow_field, self.turbine, self.grid, self.wake)
        else:
            full_flow_sequential_solver(self.farm, self.flow_field, self.turbine, self.grid, self.wake)


    ## I/O

    @classmethod
    def from_file(cls, input_file_path: str | Path, filetype: str = None) -> Floris:
        """Creates a `Floris` instance from an input file. Must be filetype
        JSON or YAML.

        Args:
            input_file_path (str): The relative or absolute file path and name to the
                input file.
            filetype (str): The type to export: [YAML | JSON]

        Returns:
            Floris: The class object instance.
        """
        input_file_path = Path(input_file_path).resolve()
        if filetype is None:
            filetype = input_file_path.suffix.strip(".")

        with open(input_file_path) as input_file:
            if filetype.lower() in ("yml", "yaml"):
                input_dict = yaml.load(input_file, Loader=yaml.SafeLoader)
            elif filetype.lower() == "json":
                input_dict = json.load(input_file)

                # TODO: This is a temporary hack to put the turbine definition into the farm.
                # Long term, we need a strategy for handling this. The YAML file format supports
                # pointers to other data, for example.
                # input_dict["farm"]["turbine"] = input_dict["turbine"]
                # input_dict.pop("turbine")
            else:
                raise ValueError("Supported import filetypes are JSON and YAML")
        return Floris.from_dict(input_dict)

    def to_file(self, output_file_path: str, filetype: str="YAML") -> None:
        """Converts the `Floris` object to an input-ready JSON or YAML file at `output_file_path`.

        Args:
            output_file_path (str): The full path and filename for where to save the file.
            filetype (str): The type to export: [YAML | JSON]
        """
        with open(output_file_path, "w+") as f:
            if filetype.lower() == "yaml":
                yaml.dump(self.as_dict(), f, default_flow_style=False)
            elif filetype.lower() == "json":
                json.dump(self.as_dict(), f, indent=2, sort_keys=False)
            else:
                raise ValueError("Supported export filetypes are JSON and YAML")



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt


class PlotDefaults:
    """
    This class sets journal-ready styles for plots.
    """

    def __init__(self):

        sns.set_style("ticks")
        sns.set_context("paper", font_scale=1.5)

        # color palette from colorbrewer (up to 4 colors, good for print and black&white printing)
        # color_brewer_palette = ['#e66101', '#5e3c99', '#fdb863', '#b2abd2']

        # most journals: 300dpi
        plt.rcParams["savefig.dpi"] = 300

        # most journals: 9 cm (or 3.5 inch) for single column width and 18.5 cm (or 7.3 inch) for double column width.
        plt.rcParams["figure.autolayout"] = False
        plt.rcParams["figure.figsize"] = 7.3, 4
        plt.rcParams["axes.labelsize"] = 16
        plt.rcParams["axes.titlesize"] = 16
        plt.rcParams["xtick.labelsize"] = 16
        plt.rcParams["ytick.labelsize"] = 16
        plt.rcParams["font.size"] = 32
        plt.rcParams["lines.linewidth"] = 2.0
        plt.rcParams["lines.markersize"] = 8
        plt.rcParams["legend.fontsize"] = 14


def data_plot(
    x,
    y,
    color="b",
    label="_nolegend_",
    x_bins=None,
    x_radius=None,
    ax=None,
    show_scatter=True,
    show_bin_points=True,
    show_confidence=True,
    min_vals=1,
    seaborn=False,
    show_80=False,
):
    """
    Plot data to a single axis (no subfigrures). Method includes flags
    to provide additional statistical context in plot (e.g. scatter,
    confidnece, etc.)

    Args:
        x (np.array): abscissa data.
        y (np.array): ordinate data.
        color (str, optional): line color.
            Defaults to 'b'.
        label (str, optional): line label used in legend.
            Defaults to '_nolegend_'.
        x_bins (np.array, optional): bin limits for abscissa data.
            Defaults to None.
        x_radius (float, optional): bin width.
            Defaults to None.
        ax (:py:class:`matplotlib.pyplot.axes`, optional):
            axes handle for plotting. Defaults to None.
        show_scatter (bool, optional): flag to control scatter plot.
            Defaults to True.
        show_bin_points (bool, optional): flag to control plot of bins.
            Defaults to True.
        show_confidence (bool, optional): flag to control plot of
            confidence interval. Defaults to True.
        min_vals (int, optional): minimum number of values required to
            merit plotting. Defaults to 1.
        seaborn (bool, optional): flag to control plotting library.
            Defaults to False.
        show_80 (bool, optional): flag to control plot of points above
            the 80th percentile. Defaults to False.
            #TODO generalize to show_percentile?

    Returns:
    Only returns values if `show_confidence` flag is active (True),
    otherwise returns (np.nan).

        x_bins (np.array): bin limits
        median_vals (np.array): median values of data in bins
        lower (np.array): lower limit of data in bins
        upper (np.array): upper limit of data in bins
    """
    if (not ax) and (not seaborn):
        fig, ax = plt.subplots()

    if seaborn:
        show_bin_points = False
        show_scatter = False

    df = pd.DataFrame({"x": x, "y": y})

    if df.shape[0] > 0:

        # If bins not provided, just use ints
        if x_bins is None:
            x_bins = np.arange(df["x"].astype(int).min(), df["x"].astype(int).max(), 1)

        # if no radius provided, use bins to determine
        if x_radius is None:
            x_radius = (x_bins[1] - x_bins[0]) / 2.0

        # now loop over bins and determine stats
        median_vals = np.zeros_like(x_bins) * np.nan
        # median_vals = np.zeros_like(x_bins) * np.nan
        count_vals = np.zeros_like(x_bins) * np.nan
        lower = np.zeros_like(x_bins) * np.nan
        upper = np.zeros_like(x_bins) * np.nan
        vals_80_up = np.zeros_like(x_bins) * np.nan
        vals_80_down = np.zeros_like(x_bins) * np.nan
        # p_down_vals = np.zeros_like(x_bins) * np.nan

        for x_idx, x_cent in enumerate(x_bins):

            df_sub = df[(df.x >= x_cent - x_radius) & (df.x <= x_cent + x_radius)]

            # TODO this conditional statement contains a lot of stuff to be cleaned up. Why all the commented content?
            if df_sub.shape[0] > min_vals:

                # Get statistics via bootstrapping
                n_bs = 40
                boot_frac = 1.0
                med_array = np.zeros(n_bs)
                for i_bs in range(n_bs):
                    # Random subset the df
                    df_rand = df_sub.sample(frac=boot_frac, replace=True)
                    # med_array[i_bs] = np.median(df_rand.y)
                    med_array[i_bs] = np.mean(df_rand.y)

                # median_vals[x_idx] = np.nanmedian(df_sub.y)
                median_vals[x_idx] = np.mean(df_sub.y)
                vals_80_down[x_idx], vals_80_up[x_idx] = np.percentile(
                    df_sub.y, [50 + 0.5 * 80.0, 50 - 0.5 * 80.0]
                )
                # mean_vals[x_idx] = np.median(ratio_array)
                count_vals[x_idx] = df_sub.shape[0]
                # ci_vals[x_idx] = scipy.stats.sem(ratio_array, ddof=1) * 1.96 # df_sub.y.apply(lambda x: scipy.stats.sem(x, ddof=1) * 1.96)
                # p_up_vals[x_idx] = p_up_func(ratio_array)# df_sub.y.apply(p_up_func)
                # p_down_vals[x_idx] = p_down_func(ratio_array)#df_sub.y.apply(p_down_func)
                # Get the confidence bounds
                confidence = 95
                conf_bounds = [50 + 0.5 * confidence, 50 - 0.5 * confidence]
                # lower[x_idx], upper[x_idx] = (2*med_vals[x_idx]-np.percentile(med_array, conf_bounds))
                lower[x_idx], upper[x_idx] = np.percentile(med_array, conf_bounds)

        # # Plot the underlying points
        if show_scatter:
            ax.scatter(
                df["x"],
                df["y"],
                color=color,
                label="_nolegend_",
                alpha=1.0,
                s=35,
                marker=".",
            )
        if show_bin_points:
            ax.scatter(
                x_bins,
                median_vals,
                color=color,
                s=count_vals,
                label="_nolegend_",
                alpha=0.6,
                marker="s",
            )
        if show_80:
            ax.plot(x_bins, vals_80_down, "--", color=color, label="_nolegend_")
            ax.plot(x_bins, vals_80_up, "--", color=color, label="_nolegend_")

        # Plot the main trend
        if not seaborn:
            ax.plot(x_bins, median_vals, label=label, color=color)
        else:
            plt.plot(x_bins, median_vals, label=label, color=color)

        if show_confidence:
            if not seaborn:
                ax.fill_between(
                    x_bins, lower, upper, alpha=0.2, color=color, label="_nolegend_"
                )
            else:
                plt.fill_between(
                    x_bins, lower, upper, alpha=0.2, color=color, label="_nolegend_"
                )

        return x_bins, median_vals, lower, upper

    else:
        ax.plot(0, 0, label=label, color=color)

        return np.nan, np.nan, np.nan, np.nan


def stacked_plot(x, groups, x_bins, ax, color_array=None):
    """
    Plot stacked histograms of data according to specified groups.

    Args:
        x (np.array): abscissa data.
        groups (list): groups of data provided by pd.Groupby()
            #TODO right?
        x_bins (np.array, optional): bin limits for abscissa data.
            Defaults to None.
        ax (:py:class:`matplotlib.pyplot.axes`, optional):
            axes handle for plotting. Defaults to None.
        color_array (list, optional): list of color specifiers.
            Defaults to None.
    """

    x_radius = (x_bins[1] - x_bins[0]) / 2.0

    # ind = np.arange(len(x_bins))

    group_vals = np.unique(groups)
    num_groups = len(group_vals)

    p_array = np.zeros((num_groups, len(x_bins)))

    for x_idx, x_cent in enumerate(x_bins):

        x_mask = (x >= x_cent - x_radius) & (x < x_cent + x_radius)

        # y_bin = y[x_mask]
        g_bin = groups[x_mask]
        num_points = len(g_bin)

        if num_points > 0:
            for g_idx, g in enumerate(group_vals):
                p_array[g_idx, x_idx] = np.sum(g_bin == g)  # / float(num_points)
    p = list()

    if color_array is not None:
        p.append(
            ax.bar(x_bins, p_array[0, :], width=x_radius * 1.5, color=color_array[0])
        )
    else:
        p.append(ax.bar(x_bins, p_array[0, :], width=x_radius * 1.5))

    for g_idx in range(1, num_groups):
        if color_array is not None:
            p.append(
                ax.bar(
                    x_bins,
                    p_array[g_idx, :],
                    bottom=p_array[g_idx - 1, :],
                    width=x_radius * 1.5,
                    color=color_array[g_idx],
                )
            )
        else:
            p.append(
                ax.bar(
                    x_bins,
                    p_array[g_idx, :],
                    bottom=p_array[g_idx - 1, :],
                    width=x_radius * 1.5,
                )
            )
    # ax.set_xticks(ind,x_bins)
    ax.legend(group_vals)
    # return group_vals,p_array


def stacked_percent_plot(x, groups, x_bins, ax, color_array=None):
    """
    Plot stacked percentage plot (normalized stacked histogram)
    according to specified groups.

    Args:
        x (np.array): abscissa data.
        groups (list): groups of data provided by pd.Groupby()
            #TODO right?
        x_bins (np.array, optional): bin limits for abscissa data.
            Defaults to None.
        ax (:py:class:`matplotlib.pyplot.axes`, optional):
            axes handle for plotting. Defaults to None.
        color_array (list, optional): list of color specifiers.
            Defaults to None.
    """
    x_radius = (x_bins[1] - x_bins[0]) / 2.0
    # ind = np.arange(len(x_bins))

    group_vals = np.unique(groups)
    num_groups = len(group_vals)

    p_array = np.zeros((num_groups, len(x_bins)))

    for x_idx, x_cent in enumerate(x_bins):

        x_mask = (x >= x_cent - x_radius) & (x < x_cent + x_radius)

        # y_bin = y[x_mask]
        g_bin = groups[x_mask]
        num_points = len(g_bin)

        if num_points > 0:
            for g_idx, g in enumerate(group_vals):
                p_array[g_idx, x_idx] = np.sum(g_bin == g) / float(num_points)
    p = list()

    if color_array is not None:
        p.append(
            ax.bar(x_bins, p_array[0, :], width=x_radius * 1.5, color=color_array[0])
        )
    else:
        p.append(ax.bar(x_bins, p_array[0, :], width=x_radius * 1.5))
    for g_idx in range(1, num_groups):
        if color_array is not None:
            p.append(
                ax.bar(
                    x_bins,
                    p_array[g_idx, :],
                    bottom=p_array[g_idx - 1, :],
                    width=x_radius * 1.5,
                    color=color_array[g_idx],
                )
            )
        else:
            p.append(
                ax.bar(
                    x_bins,
                    p_array[g_idx, :],
                    bottom=p_array[g_idx - 1, :],
                    width=x_radius * 1.5,
                )
            )
    # ax.set_xticks(ind,x_bins)
    ax.legend(group_vals, bbox_to_anchor=(1.0, 1.0))
    # return group_vals,p_array



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import attrs
from attrs import define, field

from floris.simulation import BaseClass, BaseModel
from floris.simulation.wake_deflection import (
    GaussVelocityDeflection,
    JimenezVelocityDeflection,
)
from floris.simulation.wake_combination import FLS, MAX, SOSFS
from floris.simulation.wake_turbulence.crespo_hernandez import CrespoHernandez
from floris.simulation.wake_velocity import (
    CumulativeGaussCurlVelocityDeficit,
    GaussVelocityDeficit,
    JensenVelocityDeficit
)


MODEL_MAP = {
    "combination_model": {
        "fls": FLS,
        "max": MAX,
        "sosfs": SOSFS
    },
    "deflection_model": {
        "jimenez": JimenezVelocityDeflection,
        "gauss": GaussVelocityDeflection
    },
    "turbulence_model": {
        "crespo_hernandez": CrespoHernandez
    },
    "velocity_model": {
        "cc": CumulativeGaussCurlVelocityDeficit,
        "gauss": GaussVelocityDeficit,
        "jensen": JensenVelocityDeficit
    },
}


@define
class WakeModelManager(BaseClass):
    """
    WakeModelManager is a container class for the wake velocity, deflection,
    turbulence, and combination models.

    Args:
        wake (:obj:`dict`): The wake's properties input dictionary
            - velocity_model (str): The name of the velocity model to be instantiated.
            - turbulence_model (str): The name of the turbulence model to be instantiated.
            - deflection_model (str): The name of the deflection model to be instantiated.
            - combination_model (str): The name of the combination model to be instantiated.
    """
    model_strings: dict = field(converter=dict)
    enable_secondary_steering: bool = field(converter=bool)
    enable_yaw_added_recovery: bool = field(converter=bool)
    enable_transverse_velocities: bool = field(converter=bool)

    wake_deflection_parameters: dict = field(converter=dict)
    wake_turbulence_parameters: dict = field(converter=dict)
    wake_velocity_parameters: dict = field(converter=dict, default={})

    combination_model: BaseModel = field(init=False)
    deflection_model: BaseModel = field(init=False)
    turbulence_model: BaseModel = field(init=False)
    velocity_model: BaseModel = field(init=False)

    def __attrs_post_init__(self) -> None:
        model: BaseModel = MODEL_MAP["velocity_model"][self.model_strings["velocity_model"]]
        model_parameters = self.wake_velocity_parameters[self.model_strings["velocity_model"]]
        if model_parameters is None:
            # Use model defaults
            self.velocity_model = model()
        else:
            self.velocity_model = model.from_dict(model_parameters)

        model: BaseModel = MODEL_MAP["deflection_model"][self.model_strings["deflection_model"]]
        model_parameters = self.wake_deflection_parameters[self.model_strings["deflection_model"]]
        if model_parameters is None:
            self.deflection_model = model()
        else:
            self.deflection_model = model.from_dict(model_parameters)

        model: BaseModel = MODEL_MAP["turbulence_model"][self.model_strings["turbulence_model"]]
        model_parameters = self.wake_turbulence_parameters[self.model_strings["turbulence_model"]]
        if model_parameters is None:
            self.turbulence_model = model()
        else:
            self.turbulence_model = model.from_dict(model_parameters)

        model: BaseModel = MODEL_MAP["combination_model"][self.model_strings["combination_model"]]
        self.combination_model = model()

    @model_strings.validator
    def validate_model_strings(self, instance: attrs.Attribute, value: dict) -> None:
        required_strings = [
            "velocity_model",
            "deflection_model",
            "combination_model",
            "turbulence_model"
        ]
        # Check that all required strings are given
        for s in required_strings:
            if s not in value.keys():
                raise KeyError(f"Wake: '{s}' not provided in the input but it is required.")

        # Check that no other strings are given
        for k in value.keys():
            if k not in required_strings:
                raise KeyError(f"Wake: '{k}' was given as input but it is not a valid option. Required inputs are: {', '.join(required_strings)}")

    @property
    def deflection_function(self):
        return self.deflection_model.function

    @property
    def velocity_function(self):
        return self.velocity_model.function

    @property
    def turbulence_function(self):
        return self.turbulence_model.function

    @property
    def combination_function(self):
        return self.combination_model.function



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import os
import pickle

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from floris.utilities import wrap_180


# TODO: organize by private and public methods


class PowerRose:
    """
    The PowerRose class is used to organize information about wind farm power
    production for different wind conditions (e.g., wind speed, wind direction)
    along with their frequencies of occurance to calculate the resulting annual
    energy production (AEP). Power production and AEP are considered for
    baseline operation, ideal operation without wake losses, and optionally
    optimal operation with wake steering. The primary purpose of the PowerRose
    class is for visualizing and reporting energy production and energy gains
    from wake steering. A PowerRose object can be populated with user-specified
    wind rose and power data (for example, using a :py:class:`~.tools
    WindRose` object) or data from a previously saved PowerRose object can be
    loaded.
    """

    def __init__(self,):
        """
        Instantiate a PowerRose object. No explicit arguments required, and an
        additional method will need to be called to populate the PowerRose
        object with data.
        """

    def load(self, filename):
        """
        This method loads data from a previously saved PowerRose pickle file
        into a PowerRose object.

        Args:
            filename (str): Path and filename of pickle file to load.
        """

        # self.name, self.df_power, self.df_yaw, self.df_turbine_power_no_wake, self.df_turbine_power_baseline, self.df_turbine_power_opt, self.df_combine = pickle.load(
        #    open(filename, "rb"))

        (
            self.name,
            self.df_windrose,
            self.power_no_wake,
            self.power_baseline,
            self.power_opt,
            self.use_opt,
        ) = pickle.load(open(filename, "rb"))

        # Compute energies
        self.df_power = pd.DataFrame(
            {"wd": self.df_windrose["wd"], "ws": self.df_windrose["ws"]}
        )
        self._compute_energy()

        # Compute totals
        self._compute_totals()

    def save(self, filename):
        """
        This method saves PowerRose data as a pickle file so that it can be
        imported into a PowerRose object later.

        Args:
            filename (str): Path and filename of pickle file to save.
        """
        pickle.dump(
            [
                self.name,
                self.df_windrose,
                self.power_no_wake,
                self.power_baseline,
                self.power_opt,
                self.use_opt,
            ],
            open(filename, "wb"),
        )

    # def _all_combine(self):
    #     df_power = self.df_power.copy(deep=True)
    #     df_yaw = self.df_yaw.copy(deep=True)
    #     df_turbine_power_no_wake = self.df_turbine_power_no_wake.copy(
    #         deep=True)
    #     df_turbine_power_baseline = self.df_turbine_power_baseline.copy(
    #         deep=True)
    #     df_turbine_power_opt = self.df_turbine_power_opt.copy(deep=True)

    #     # Adjust the column names for uniqunes
    #     df_yaw.columns = [
    #         'yaw_%d' % c if type(c) is int else c for c in df_yaw.columns
    #     ]
    #     df_turbine_power_no_wake.columns = [
    #         'tnw_%d' % c if type(c) is int else c
    #         for c in df_turbine_power_no_wake.columns
    #     ]
    #     df_turbine_power_baseline.columns = [
    #         'tb_%d' % c if type(c) is int else c
    #         for c in df_turbine_power_baseline.columns
    #     ]
    #     df_turbine_power_opt.columns = [
    #         'topt_%d' % c if type(c) is int else c
    #         for c in df_turbine_power_opt.columns
    #     ]

    #     # Merge
    #     df_combine = df_power.merge(df_yaw, on=['ws', 'wd'])
    #     df_combine = df_combine.merge(df_turbine_power_no_wake,
    #                                   on=['ws', 'wd'])
    #     df_combine = df_combine.merge(df_turbine_power_baseline,
    #                                   on=['ws', 'wd'])
    #     df_combine = df_combine.merge(df_turbine_power_opt, on=['ws', 'wd'])

    #     return df_combine

    def _norm_frequency(self, df):
        print("Norming frequency total of %.2f to 1.0" % df.freq_val.sum())
        df["freq_val"] = df.freq_val / df.freq_val.sum()
        return df

    def _compute_energy(self):
        self.df_power["energy_no_wake"] = self.df_windrose.freq_val * self.power_no_wake
        self.df_power["energy_baseline"] = (
            self.df_windrose.freq_val * self.power_baseline
        )
        if self.use_opt:
            self.df_power["energy_opt"] = self.df_windrose.freq_val * self.power_opt

    def _compute_totals(self):
        df = self.df_power.copy(deep=True)
        df = df.sum()

        # Get total annual energy amounts
        self.total_no_wake = (8760 / 1e9) * df.energy_no_wake
        self.total_baseline = (8760 / 1e9) * df.energy_baseline
        if self.use_opt:
            self.total_opt = (8760 / 1e9) * df.energy_opt

        # Get wake loss amounts
        self.baseline_percent = self.total_baseline / self.total_no_wake
        self.baseline_wake_loss = 1 - self.baseline_percent

        if self.use_opt:
            self.opt_percent = self.total_opt / self.total_no_wake
            self.opt_wake_loss = 1 - self.opt_percent

        # Percent gain
        if self.use_opt:
            self.percent_gain = (
                self.total_opt - self.total_baseline
            ) / self.total_baseline
            self.reduction_in_wake_loss = (
                -1
                * (self.opt_wake_loss - self.baseline_wake_loss)
                / self.baseline_wake_loss
            )

    def make_power_rose_from_user_data(
        self, name, df_windrose, power_no_wake, power_baseline, power_opt=None
    ):
        """
        This method populates the PowerRose object with a user-specified wind
        rose containing wind direction, wind speed, and additional optional
        variables, as well as baseline wind farm power, ideal wind farm power
        without wake losses, and optionally optimal wind farm power with wake
        steering corresponding to each wind condition.

        TODO: Add inputs for turbine-level power and optimal yaw offsets.

        Args:
            name (str): The name of the PowerRose object.
            df_windrose (pandas.DataFrame): A DataFrame with wind rose
                information containing at least
                the following columns:

                - **wd** (*float*) - Wind direction bin center values (deg).
                - **ws** (*float*) - Wind speed bin center values (m/s).
                - **freq_val** (*float*) - The frequency of occurance of the
                wind conditions in the other columns.

            power_no_wake (iterable): A list of wind farm power without wake
                losses corresponding to the wind conditions in df_windrose (W).
            power_baseline (iterable): A list of baseline wind farm power with
                wake losses corresponding to the wind conditions in df_windrose
                (W).
            power_opt (iterable, optional): A list of optimal wind farm power
                with wake steering corresponding to the wind conditions in
                df_windrose (W). Defaults to None.
        """
        self.name = name
        if df_windrose is not None:
            self.df_windrose = self._norm_frequency(df_windrose)
        self.power_no_wake = power_no_wake
        self.power_baseline = power_baseline
        self.power_opt = power_opt

        # Only use_opt data if provided
        if power_opt is None:
            self.use_opt = False
        else:
            self.use_opt = True

        # # Make a single combined frame in case it's useful (Set aside for now)
        # self.df_combine = self._all_combine()

        # Compute energies
        self.df_power = pd.DataFrame({"wd": df_windrose["wd"], "ws": df_windrose["ws"]})
        self._compute_energy()

        # Compute totals
        self._compute_totals()

    def report(self):
        """
        This method prints information about annual energy production (AEP)
        using the PowerRose object data. The AEP in GWh is listed for ideal
        operation without wake losses, baseline operation, and optimal
        operation with wake steering, if optimal power data are stored. The
        wind farm efficiency (% of ideal energy production) and wake loss
        percentages are listed for baseline and optimal operation (if optimal
        power is stored), along with the AEP gain from wake steering (again, if
        optimal power is stored). The AEP gain from wake steering is also
        listed as a percentage of wake losses recovered, if applicable.
        """
        if self.use_opt:
            print("=============================================")
            print("Case %s has results:" % self.name)
            print("=============================================")
            print("-\tNo-Wake\t\tBaseline\tOpt ")
            print("---------------------------------------------")
            print(
                "AEP (GWh)\t%.1E\t\t%.1E\t\t%.1E"
                % (self.total_no_wake, self.total_baseline, self.total_opt)
            )
            print(
                "%%\t--\t\t%.1f%%\t\t%.1f%%"
                % (100.0 * self.baseline_percent, 100.0 * self.opt_percent)
            )
            print(
                "Wk Loss\t--\t\t%.1f%%\t\t%.1f%%"
                % (100.0 * self.baseline_wake_loss, 100.0 * self.opt_wake_loss)
            )
            print("AEP Gain --\t\t--\t\t%.1f%%" % (100.0 * self.percent_gain))
            print("Loss Red --\t\t--\t\t%.1f%%" % (100.0 * self.reduction_in_wake_loss))
        else:
            print("=============================================")
            print("Case %s has results:" % self.name)
            print("=============================================")
            print("-\tNo-Wake\t\tBaseline ")
            print("---------------------------------------------")
            print("AEP (GWh)\t%.1E\t\t%.1E" % (self.total_no_wake, self.total_baseline))
            print("%%\t--\t\t%.1f%%" % (100.0 * self.baseline_percent))
            print("Wk Loss\t--\t\t%.1f%%" % (100.0 * self.baseline_wake_loss))

    def plot_by_direction(self, axarr=None):
        """
        This method plots energy production, wind farm efficiency, and energy
        gains from wake steering (if applicable) as a function of wind
        direction. If axes are not provided, new ones are created. The plots
        include:

        1) The energy production as a function of wind direction for the
        baseline and, if applicable, optimal wake steering cases normalized by
        the maximum energy production.
        2) The wind farm efficiency (energy production relative to energy
        production without wake losses) as a function of wind direction for the
        baseline and, if applicable, optimal wake steering cases.
        3) Percent gain in energy production with optimal wake steering as a
        function of wind direction. This third plot is only created if optimal
        power data are stored in the PowerRose object.

        Args:
            axarr (numpy.ndarray, optional): An array of 2 or 3
                :py:class:`matplotlib.axes._subplots.AxesSubplot` axes objects
                on which data are plotted. Three axes are rquired if the
                PowerRose object contains optimal power data. Default is None.

        Returns:
            numpy.ndarray: An array of 2 or 3
            :py:class:`matplotlib.axes._subplots.AxesSubplot` axes objects on
            which the data are plotted.
        """

        df = self.df_power.copy(deep=True)
        df = df.groupby("wd").sum().reset_index()

        if self.use_opt:

            if axarr is None:
                fig, axarr = plt.subplots(3, 1, sharex=True)

            ax = axarr[0]
            ax.plot(
                df.wd,
                df.energy_baseline / np.max(df.energy_opt),
                label="Baseline",
                color="k",
            )
            ax.axhline(
                np.mean(df.energy_baseline / np.max(df.energy_opt)), color="r", ls="--"
            )
            ax.plot(
                df.wd,
                df.energy_opt / np.max(df.energy_opt),
                label="Optimized",
                color="r",
            )
            ax.axhline(
                np.mean(df.energy_opt / np.max(df.energy_opt)), color="r", ls="--"
            )
            ax.set_ylabel("Normalized Energy")
            ax.grid(True)
            ax.legend()
            ax.set_title(self.name)

            ax = axarr[1]
            ax.plot(
                df.wd,
                df.energy_baseline / df.energy_no_wake,
                label="Baseline",
                color="k",
            )
            ax.axhline(
                np.mean(df.energy_baseline) / np.mean(df.energy_no_wake),
                color="k",
                ls="--",
            )
            ax.plot(
                df.wd, df.energy_opt / df.energy_no_wake, label="Optimized", color="r"
            )
            ax.axhline(
                np.mean(df.energy_opt) / np.mean(df.energy_no_wake), color="r", ls="--"
            )
            ax.set_ylabel("Wind Farm Efficiency")
            ax.grid(True)
            ax.legend()

            ax = axarr[2]
            ax.plot(
                df.wd,
                100.0 * (df.energy_opt - df.energy_baseline) / df.energy_baseline,
                "r",
            )
            ax.axhline(
                100.0
                * (df.energy_opt.mean() - df.energy_baseline.mean())
                / df.energy_baseline.mean(),
                df.energy_baseline.mean(),
                color="r",
                ls="--",
            )
            ax.set_ylabel("Percent Gain")
            ax.set_xlabel("Wind Direction (deg)")

            return axarr

        else:

            if axarr is None:
                fig, axarr = plt.subplots(2, 1, sharex=True)

            ax = axarr[0]
            ax.plot(
                df.wd,
                df.energy_baseline / np.max(df.energy_baseline),
                label="Baseline",
                color="k",
            )
            ax.axhline(
                np.mean(df.energy_baseline / np.max(df.energy_baseline)),
                color="r",
                ls="--",
            )
            ax.set_ylabel("Normalized Energy")
            ax.grid(True)
            ax.legend()
            ax.set_title(self.name)

            ax = axarr[1]
            ax.plot(
                df.wd,
                df.energy_baseline / df.energy_no_wake,
                label="Baseline",
                color="k",
            )
            ax.axhline(
                np.mean(df.energy_baseline) / np.mean(df.energy_no_wake),
                color="k",
                ls="--",
            )
            ax.set_ylabel("Wind Farm Efficiency")
            ax.grid(True)
            ax.legend()

            ax.set_xlabel("Wind Direction (deg)")

            return axarr

    # def wake_loss_at_direction(self, wd):
    #     """
    #     Calculate wake losses for a given direction. Plot rose figures
    #     for Power, Energy, Baseline power, Optimal gain, Total Gain,
    #     Percent Gain, etc.

    #     Args:
    #         wd (float): Wind direction of interest.

    #     Returns:
    #         tuple: tuple containing:

    #             -   **fig** (*plt.figure*): Figure handle.
    #             -   **axarr** (*list*): list of axis handles.
    #     """

    #     df = self.df_power.copy(deep=True)

    #     # Choose the nearest direction
    #     # Find nearest wind direction
    #     df['dist'] = np.abs(wrap_180(df.wd - wd))
    #     wd_select = df[df.dist == df.dist.min()]['wd'].unique()[0]
    #     print('Nearest wd to %.1f is %.1f' % (wd, wd_select))
    #     df = df[df.wd == wd_select]

    #     df = df.groupby('ws').sum().reset_index()

    #     fig, axarr = plt.subplots(4, 2, sharex=True, figsize=(14, 12))

    #     ax = axarr[0, 0]
    #     ax.set_title('Power')
    #     ax.plot(df.ws, df.power_no_wake, 'k', label='No Wake')
    #     ax.plot(df.ws, df.power_baseline, 'b', label='Baseline')
    #     ax.plot(df.ws, df.power_opt, 'r', label='Opt')
    #     ax.set_ylabel('Total')
    #     ax.grid()

    #     ax = axarr[0, 1]
    #     ax.set_title('Energy')
    #     ax.plot(df.ws, df.energy_no_wake, 'k', label='No Wake')
    #     ax.plot(df.ws, df.energy_baseline, 'b', label='Baseline')
    #     ax.plot(df.ws, df.energy_opt, 'r', label='Opt')
    #     ax.legend()
    #     ax.grid()

    #     ax = axarr[1, 0]
    #     ax.plot(df.ws,
    #             df.power_baseline / df.power_no_wake,
    #             'b',
    #             label='Baseline')
    #     ax.plot(df.ws, df.power_opt / df.power_no_wake, 'r', label='Opt')
    #     ax.set_ylabel('Percent')
    #     ax.grid()

    #     ax = axarr[1, 1]
    #     ax.plot(df.ws,
    #             df.energy_baseline / df.energy_no_wake,
    #             'b',
    #             label='Baseline')
    #     ax.plot(df.ws, df.energy_opt / df.energy_no_wake, 'r', label='Opt')
    #     ax.grid()

    #     ax = axarr[2, 0]
    #     ax.plot(df.ws, (df.power_opt - df.power_baseline), 'r')
    #     ax.set_ylabel('Total Gain')
    #     ax.grid()

    #     ax = axarr[2, 1]
    #     ax.plot(df.ws, (df.energy_opt - df.energy_baseline), 'r')
    #     ax.grid()

    #     ax = axarr[3, 0]
    #     ax.plot(df.ws, (df.power_opt - df.power_baseline) / df.power_baseline,
    #             'r')
    #     ax.set_ylabel('Percent Gain')
    #     ax.grid()
    #     ax.set_xlabel('Wind Speed (m/s)')

    #     ax = axarr[3, 1]
    #     ax.plot(df.ws,
    #             (df.energy_opt - df.energy_baseline) / df.energy_baseline, 'r')
    #     ax.grid()
    #     ax.set_xlabel('Wind Speed (m/s)')

    #     return fig, axarr
